#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.numerical` module provides an abstract class
:py:class:`~.BaseNumerical` that implements basic mathematical operations. This
is used by series and grid functions.

To work, class derived from :py:class:`~.BaseNumerical` have to implement
methods to describe (1) how to apply a function that takes the object as argument,
(2) how to apply a function that takes two objects as arguments, and (3) how to
apply a reduction.

"""

from abc import ABC, abstractmethod

import numpy as np


# Note, we test this class testing its derived class TimeSeries
class BaseNumerical(ABC):
    """Base abstract class for generic numerical data.

    This class provides the infrastructure needed to implement mathematical
    operations in all the derived classes.

    The derived classes have to implement:
    - _apply_unary(self, function, *args, **kwargs)
       that returns function(self, *args, **kwargs)
    - _apply_binary(self, other, function, *args, **kwargs)
       that returns function(self, other, *args, **kwargs)
    - _apply_reduction(self, function, *args, **kwargs)
       that returns function(self, *args, **kwargs)
    - _apply_to_self(self, function, *args, **kwargs)
       that returns applies function(self, *args, **kwargs)
       and modifies self.

    """

    @abstractmethod
    def _apply_unary(self, function, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def _apply_binary(self, other, function, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def _apply_reduction(self, reduction, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def _apply_to_self(self, function, *args, **kwargs):
        raise NotImplementedError

    def __add__(self, other):
        return self._apply_binary(other, np.add)

    __radd__ = __add__

    def __sub__(self, other):
        return self._apply_binary(other, np.subtract)

    def __rsub__(self, other):
        return -self._apply_binary(other, np.subtract)

    def __mul__(self, other):
        return self._apply_binary(other, np.multiply)

    __rmul__ = __mul__

    def __truediv__(self, other):
        if other == 0:
            raise ValueError("Cannot divide by zero")
        return self._apply_binary(other, np.divide)

    def __rtruediv__(self, other):
        # This self._apply_binary(other, np.divide)
        # divives self by other, so, we reverse that
        # with ** -1
        return (self._apply_binary(other, np.divide)) ** -1

    def __pow__(self, other):
        return self._apply_binary(other, np.power)

    def __iadd__(self, other):
        return self + other

    def __isub__(self, other):
        return self - other

    def __imul__(self, other):
        return self * other

    def __itruediv__(self, other):
        return self / other

    def __ipow__(self, other):
        return self**other

    def __neg__(self):
        return self._apply_unary(np.negative)

    def __abs__(self):
        return self._apply_unary(np.abs)

    def min(self):
        return self._apply_reduction(np.min)

    def max(self):
        return self._apply_reduction(np.max)

    def mean(self):
        return self._apply_reduction(np.mean)

    average = mean

    def median(self):
        return self._apply_reduction(np.median)

    def std(self):
        return self._apply_reduction(np.std)

    def nanmin(self):
        return self._apply_reduction(np.nanmin)

    def nanmax(self):
        return self._apply_reduction(np.nanmax)

    def abs_min(self):
        """Return the minimum of the absolute value"""
        # skipcq PYL-W0212
        return abs(self)._apply_reduction(np.min)

    def abs_max(self):
        """Return the maximum of the absolute value"""
        # skipcq PYL-W0212
        return abs(self)._apply_reduction(np.max)

    def abs_nanmin(self):
        """Return the minimum of the absolute value ignoring NaNs"""
        # skipcq PYL-W0212
        return abs(self)._apply_reduction(np.nanmin)

    def abs_nanmax(self):
        """Return the maximum of the absolute value ignoring NaNs"""
        # skipcq PYL-W0212
        return abs(self)._apply_reduction(np.nanmax)

    def abs(self):
        return self._apply_unary(np.abs)

    def real(self):
        return self._apply_unary(np.real)

    def imag(self):
        return self._apply_unary(np.imag)

    def sin(self):
        return self._apply_unary(np.sin)

    def cos(self):
        return self._apply_unary(np.cos)

    def tan(self):
        return self._apply_unary(np.tan)

    def arcsin(self):
        return self._apply_unary(np.arcsin)

    def arccos(self):
        return self._apply_unary(np.arccos)

    def arctan(self):
        return self._apply_unary(np.arctan)

    def sinh(self):
        return self._apply_unary(np.sinh)

    def cosh(self):
        return self._apply_unary(np.cosh)

    def tanh(self):
        return self._apply_unary(np.tanh)

    def arcsinh(self):
        return self._apply_unary(np.arcsinh)

    def arccosh(self):
        return self._apply_unary(np.arccosh)

    def arctanh(self):
        return self._apply_unary(np.arctanh)

    def sqrt(self):
        return self._apply_unary(np.sqrt)

    def exp(self):
        return self._apply_unary(np.exp)

    def log(self):
        return self._apply_unary(np.log)

    def log2(self):
        return self._apply_unary(np.log2)

    def log10(self):
        return self._apply_unary(np.log10)

    def conjugate(self):
        return self._apply_unary(np.conjugate)

    # Masked functions

    def ma_sqrt(self):
        return self._apply_unary(np.ma.sqrt)

    def ma_exp(self):
        return self._apply_unary(np.ma.exp)

    def ma_log(self):
        return self._apply_unary(np.ma.log)

    def ma_log2(self):
        return self._apply_unary(np.ma.log2)

    def ma_log10(self):
        return self._apply_unary(np.ma.log10)

    def ma_sin(self):
        return self._apply_unary(np.ma.sin)

    def ma_cos(self):
        return self._apply_unary(np.ma.cos)

    def ma_tan(self):
        return self._apply_unary(np.ma.tan)

    def ma_arcsin(self):
        return self._apply_unary(np.ma.arcsin)

    def ma_arccos(self):
        return self._apply_unary(np.ma.arccos)

    def ma_arctan(self):
        return self._apply_unary(np.ma.arctan)

    def ma_arccosh(self):
        return self._apply_unary(np.ma.arccosh)

    def ma_arctanh(self):
        return self._apply_unary(np.ma.arctanh)

    # Create masks

    def masked_equal(self, value):
        """Return a new objected masked where data is equal to given value."""
        return self._apply_unary(np.ma.masked_equal, value)

    def mask_equal(self, value):
        """Mask where data is equal to given value."""
        self._apply_to_self(self.masked_equal, value)

    def masked_greater(self, value):
        """Return a new objected masked where data is greater to given value."""
        return self._apply_unary(np.ma.masked_greater, value)

    def mask_greater(self, value):
        """Mask where data is greater to given value."""
        self._apply_to_self(self.masked_greater, value)

    def masked_greater_equal(self, value):
        """Return a new objected masked where data is greater or equal to given value."""
        return self._apply_unary(np.ma.masked_greater_equal, value)

    def mask_greater_equal(self, value):
        """Mask where data is greater or equal to given value."""
        self._apply_to_self(self.masked_greater_equal, value)

    def masked_inside(self, value1, value2):
        """Return a new objected masked where data is inside the given values."""
        return self._apply_unary(np.ma.masked_inside, value1, value2)

    def mask_inside(self, value1, value2):
        """Mask where data is inside the given values."""
        self._apply_to_self(self.masked_inside, value1, value2)

    def masked_invalid(self):
        """Return a new objected masked where data is invalid (NaNs or infs)."""
        return self._apply_unary(np.ma.masked_invalid)

    def mask_invalid(self):
        """Mask where data is invalid (NaNs of infs)."""
        self._apply_to_self(self.masked_invalid)

    def masked_less(self, value):
        """Return a new objected masked where data is less to given value."""
        return self._apply_unary(np.ma.masked_less, value)

    def mask_less(self, value):
        """Mask where data is less to given value."""
        self._apply_to_self(self.masked_less, value)

    def masked_less_equal(self, value):
        """Return a new objected masked where data is less or equal to given value."""
        return self._apply_unary(np.ma.masked_less_equal, value)

    def mask_less_equal(self, value):
        """Mask where data is less or equal to given value."""
        self._apply_to_self(self.masked_less_equal, value)

    def masked_not_equal(self, value):
        """Return a new objected masked where data is not equal to given value."""
        return self._apply_unary(np.ma.masked_not_equal, value)

    def mask_not_equal(self, value):
        """Mask where data is not equal to given value."""
        self._apply_to_self(self.masked_not_equal, value)

    def masked_outside(self, value1, value2):
        """Return a new objected masked where data is outside the given values."""
        return self._apply_unary(np.ma.masked_outside, value1, value2)

    def mask_outside(self, value1, value2):
        """Mask where data is outside the given values."""
        self._apply_to_self(self.masked_outside, value1, value2)



#!/usr/bin/env python3

# Copyright (C) 2022-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import re

import h5py

from kuibit.simdir import SimDir


class PittNullOne:
    def __init__(self, files: list[str]) -> None:
        ...

        # We read off all the metadata to ensure that they are consistent. To do
        # so, we put them in sets and see how many elements they end up having
        Rin: set[float] = set()
        Rout: set[float] = set()
        dim: set[tuple[int, int]] = set()
        spin: set[int] = set()

        for f in files:
            with h5py.File(f) as h5f:
                # We unpack the metadata because they are saved as arrays
                Rin.add(*h5f["metadata"].attrs["Rin"])
                Rout.add(*h5f["metadata"].attrs["Rout"])

                # We make the array hashable so that we can add it to the set
                dim.add(tuple(h5f["metadata"].attrs["dim"]))

                spin.add(*h5f["metadata"].attrs["spin"])

        if len(Rin) != 1:
            raise RuntimeError(f"Multiple/no values of Rin found: {Rin}")
        if len(Rout) != 1:
            raise RuntimeError(f"Multiple/no values of Rout found: {Rout}")
        if len(dim) != 1:
            raise RuntimeError(f"Multiple/no values of dim found: {dim}")
        if len(spin) != 1:
            raise RuntimeError(f"Multiple/no values of dim found: {spin}")

        # We are clear. Now we set the metadata with set unpacking
        (self.Rin,) = Rin
        (self.Rout,) = Rout
        (self.dim,) = dim
        (self.spin,) = spin


class PittNullDir:
    def __init__(self, sd: SimDir) -> None:
        """Constructor.

        :param sd:  SimDir object providing access to data directory.
        :type sd:   SimDir

        """
        # First, we organize all the metric_obs_D_Decomp.h5 files
        #
        # _pitt_files is a dictionary whose keys are the radius indices (0, 1,
        # ...) and values the list of the files associated to that index.
        self._pitt_files: dict[int, list[str]] = {}

        rx_filename = re.compile(r"^metric_obs_(\d+)_Decomp.h5$")
        for path in sd.allfiles:
            filename = os.path.split(path)[-1]
            matched = rx_filename.search(filename)
            if matched is not None:
                rad_index = int(matched.group(1))
                self._pitt_files.setdefault(rad_index, []).append(path)

        # We cache what we have already read
        self._pitt_ones: dict[int, PittNullOne] = {}

    def __getitem__(self, rad_index: int) -> PittNullOne:
        return PittNullOne(self._pitt_files[rad_index])



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.series` module provides a base class :py:class:`~.BaseSeries`
for representing and handling series (from which :py:class:`~.TimeSeries` and
:py:class:`~.FrequencySeries` are derived).

:py:class:`~.BaseSeries` handles series that have a independent variable ``x``
and a dependent variable ``y``. The derived classes have to implement setters
and getters if they need to rename these variables (e.g. ``x -> t``). The
independent variable has to be monotonically increasing.

:py:class:`~.BaseSeries` implements several methods for operations on series.
Most of these methods are available in two flavors: those that return a new
:py:class:`~.BaseSeries`, and those which modify the object in place. The latter
have names with imperative verbs.

This module also provides the useful function :py:func:`~.sample_common`, which
takes a list of series and resamples them to their common points.

"""

import warnings
from typing import Tuple

import numpy as np
import numpy.typing as npty
from scipy import integrate, interpolate, signal

from kuibit.attr_dict import AttributeDictionary
from kuibit.numerical import BaseNumerical
from kuibit.tensor import Tensor


class _AttributeDictionaryNumPy(AttributeDictionary):
    """Class that maps a dictionary to attributes and that has a method
    ``to_numpy``.

    This is used only to fake being Pandas so that Series can be plotted with
    matplotlib. See comments in BaseSeries.

    """

    def to_numpy(self):
        # This method only makes sense in the context of faking pandas. See
        # comments in BaseSeries about this.
        return self.values


# Note, we test this class testing its derived class TimeSeries
class BaseSeries(BaseNumerical):
    """Base class (not intended for direct use) for generic series data in
    which the independendent variable x is sorted.

    This class is already rich of features.

    .. note:

        Derived class should define setters and getters to handle ``x``
        and ``y``. This is where the data is stored.

        The idea is the following. The actual data is stored in the
        ``BaseSeries` properties ``data_x`` and ``data_y``. These are
        accessible from the derived classes. However, we don't want the
        derived classes to use directly ``data_x`` and ``data_y``: they
        should use something that clearly inform the user of their meaning,
        like ``t`` or ``f`` (time or frequency). To do this, we have to
        define getters and setters that access and modify ``data_x``
        and ``y`` but use more meaningful names. To define a getters,
        simply use the ``@property`` decorator:

        .. code-block:: python

            @property
            def t(self):
                 return self.data_x

        With these, ``.t`` will return ``self.data_x``. For a setter,

        .. code-block:: python

            @t.setter
            def t(self, t):
                # This is defined BaseClass
                self.data_x = t

        This is called when with ``.t = something``. Once these are defined,
        the derived classes should use their getters and setters.

    :ivar data_x: Independent variable.
    :vartype data_x: 1D NumPy array or float
    :ivar y: Dependent variable.
    :vartype y: 1D NumPy array or float

    :ivar spline_real: Coefficients for a spline represent of the real part
                       of y.
    :vartype spline_real: Tuple
    :ivar spline_imag: Coefficients for a spline represent of the real part
                       of y.
    :vartype spline_imag: Tuple

    """

    @staticmethod
    def _make_array(x):
        """Return a NumPy array version of x (if x is not already an array)."""
        return np.atleast_1d(x) if not isinstance(x, np.ndarray) else x

    def _return_array_if_monotonic(self, x_array):
        """Return the import array if it has length 1 or if it is
        monotonically increasing. Otherwise return error.

        We assume x_array is an array. We will not check for this, it is up to
        the developer to guarantee this. If this is not true, some errors will
        be thrown.

        :param x_array: Array to check if it is monotonically increasing.
        :type x_array: 1d NumPy array

        :returns: Input array, if increasing monotonically.
        :rtype: 1d NumPy array

        """
        if len(x_array) > 1:
            # Here we compute directly the diff because it seems faster
            # than using np.diff

            # Example:
            # self.x = [1,2,3]
            # self.x[1:] = [2, 3]
            # self.x[:-1] = [1, 2]
            # dx = [1,1]
            dx = x_array[1:] - x_array[:-1]
            if dx.min() <= 0:
                # HACK: To provide more useful information we assume
                #       that the derived classes are named like TimeSeries.
                #       Then, we remove the "series"
                name = type(self).__name__
                x_name = name[:-6]
                raise ValueError(f"{x_name} not monotonically increasing")

        return x_array

    def __init__(self, x, y, guarantee_x_is_monotonic=False):
        """When guarantee_x_is_monotonic is True no checks will be perform to
        make sure that x is monotonically increasing (increasing performance).
        This should is used internally whenever a new series is returned from
        self, since we have already checked that data_x is good.

        :param x: Independent variable.
        :type x: 1d NumPy array or list
        :param y: Dependent variable.
        :type y: 1d NumPy array or list
        :param guarantee_x_is_monotonic: Whether we can skip the check on monotonicity.
        :param guarantee_x_is_monotonic: bool

        """

        x_array = self._make_array(x)
        y_array = self._make_array(y)

        if len(x_array) != len(y_array):
            raise ValueError("Data length mismatch")

        if len(x_array) == 0:
            raise ValueError("Trying to construct empty Series.")

        if not guarantee_x_is_monotonic:
            x_array = self._return_array_if_monotonic(x_array)

        # The copy is because we don't want to change the input values
        self.__data_x = x_array.copy()
        self.__data_y = y_array.copy()

        # The data is stored in the members self.data_x and self.data_y. We
        # will never access these directly. We have setters and getters to that
        # we can do stuff when variables change. For example, we want to
        # compute or update splines. The setter and getters are for x and y

        # We keep this flag around to know when we have to recompute the
        # splines. Operations that invalidate the splines MUST reset this flag
        # to True.
        self.invalid_spline = True
        # Here we also define the splines as empty objects so that we know
        # that they are attributes of the class and they are not uninitialized
        self.spline_real = None
        self.spline_imag = None

    @property
    def x(self):
        return self.__data_x

    @x.setter
    def x(self, x):
        x_array = self._make_array(x)
        if len(x_array) != len(self.x):
            raise ValueError("You cannot change the length of the series")
        x_array = self._return_array_if_monotonic(x_array)

        # This series should own the data, so we copy (to avoid accidentally
        # changing some other variable).
        # If you do self.x = z
        # and then self.x = *2
        # z will change (if we don't copy)
        self.__data_x = x_array.copy()

        # Invalidate the spline
        self.invalid_spline = True

    @property
    def y(self):
        return self.__data_y

    @y.setter
    def y(self, y):
        y_array = self._make_array(y)
        if len(y_array) != len(self.__data_y):
            raise ValueError("You cannot change the length of the series")

        # This series should own the data, so we copy (to avoid accidentally
        # changing some other variable)
        self.__data_y = y_array.copy()

        # Invalidate the spline
        self.invalid_spline = True

    # Here is where we pretend to be pandas. We want to be able to plot our
    # series with matplotlib. Unfortunately, there is no easy way to provide a
    # custom object to the plot functions. However, matplotlib has a special
    # hook for pandas in the function matplotlib.cbook.index_of. In this
    # function is checked if the index property is available, in which case,
    # index.values and values are returned. We use this to make our objects
    # plottable.
    # The function in matplotlib < 3.5.3 is:
    # try:
    #    return y.index.values, y.values
    # except AttributeError:
    #    y = _check_1d(y)
    #    return np.arange(y.shape[0], dtype=float), y
    #
    # It got updated to
    # try:
    #     return y.index.to_numpy(), y.to_numpy()
    # except AttributeError:
    #     pass
    # try:
    #     y = _check_1d(y)
    # except (np.VisibleDeprecationWarning, ValueError):
    #     # NumPy 1.19 will warn on ragged input, and we can't actually use it.
    #     pass
    # else:
    #     return np.arange(y.shape[0], dtype=float), y
    # raise ValueError('Input could not be cast to an at-least-1D NumPy array')
    #
    # Let us try to be compatible with both (luckily, we can easily do that).
    #
    # For the older version, if we provide index.values and values, we can
    # return x and y.
    #
    # For the newer version, we also have to provide the correct methods named
    # `to_numpy` to both this class and the object that is return by the index
    # attribute. So, we define a _AttributeDictionaryNumPy class that does that.

    def to_numpy(self):
        """Return the data as NumPy array. Equivalent to ``self.y``.

        This function is here to enable compatibility matplotlib.
        """
        return self.y

    @property
    def values(self):
        """Fake pandas properties, to make Series objects plottable by
        matplotlib.
        """
        return self.y

    @property
    def index(self):
        """Fake pandas properties, to make Series objects plottable by
        matplotlib.
        """
        return _AttributeDictionaryNumPy({"values": self.x})

    @property
    def mask(self):
        """Return where the data is valid (according to the mask).

        :returns: Array of True/False of the same length of the data.
                  False where the data is valid, true where is not.
        :rtype: 1D array of bool
        """
        if self.is_masked():
            return np.ma.getmask(self.y)
        return np.zeros(len(self), dtype=bool)

    @property
    def xmin(self):
        """Return the minimum of the independent variable x.

        :rvalue: Minimum of x.
        :rtype: float
        """
        return self.x[0]

    @property
    def xmax(self):
        """Return the maximum of the independent variable x.

        :rvalue: Maximum of x
        :rtype: float
        """
        return self.x[-1]

    def is_regularly_sampled(self):
        """Return whether the series is regularly sampled.

        If the series is only one point, an error is raised.

        :returns:  Is the series regularly sampled?
        :rtype:    bool
        """
        if len(self) == 1:
            raise RuntimeError(
                "Series is only one point, "
                "it does not make sense to compute dx"
            )

        dx = self.x[1:] - self.x[:-1]

        return np.allclose(dx, dx[0], atol=1e-14)

    def __len__(self):
        """The number of data points."""
        return len(self.x)

    def __iter__(self):
        for x, y in zip(self.x, self.y):
            yield x, y

    def is_complex(self):
        """Return whether the data is complex.

        :returns:  True if the data is complex, false if it is not.
        :rtype:   bool

        """
        return issubclass(self.y.dtype.type, complex)

    def is_masked(self) -> bool:
        """Return whether the x or y are masked.

        :returns:  True if the x or y are masked, false if it is not.
        :rtype:   bool

        """
        return isinstance(self.y, np.ma.MaskedArray) or isinstance(
            self.x, np.ma.MaskedArray
        )

    def x_at_maximum_y(self):
        """Return the value of x when y is maximum.

        :returns: Value of x when y is maximum.
        :rtype: float
        """
        return self.x[np.argmax(self.y)]

    def x_at_minimum_y(self):
        """Return the value of x when y is minimum.

        :returns: Value of x when y is minimum.
        :rtype: float
        """
        return self.x[np.argmin(self.y)]

    def x_at_abs_maximum_y(self):
        """Return the value of x when abs(y) is maximum.

        :returns: Value of x when abs(y) is maximum.
        :rtype: float
        """
        return self.x[np.argmax(np.abs(self.y))]

    def x_at_abs_minimum_y(self):
        """Return the value of x when abs(y) is minimum.

        :returns: Value of x when abs(y) is minimum.
        :rtype: float
        """
        return self.x[np.argmin(np.abs(self.y))]

    def _local_extrema(
        self, y: npty.NDArray, *args, include_edges: bool = True, **kwargs
    ) -> Tuple[npty.NDArray]:
        """Use SciPy's ``find_peaks`` to find the local minima and maxima.

        Unkown arguments are passed to ``find_peaks``.

        If the signal is complex, the absolute value is taken.

        If ``include_edges`` is True, the edges are considered among the
        possible extrema.

        :returns: Coordinate and value of the peaks.
        :rtype: Tuple of NumPy arrays

        """
        # TODO (FEATURE): Improve accuracy
        #
        # We can use splines or quadratic approximations to improve the location
        # the peak (ie, find it within a cell)

        offset = 0
        if include_edges:
            # https://stackoverflow.com/a/60096220
            min_y = (min(y),)
            y = np.concatenate((min_y, y, min_y))
            offset = 1

        peak_indices, _ = signal.find_peaks(y, *args, **kwargs)

        # If we included the edges, we added an extra point, se we have to
        # remove the offset
        peak_indices = peak_indices - offset
        return self.x[peak_indices], self.y[peak_indices]

    def local_maxima(
        self, *args, include_edges: bool = True, **kwargs
    ) -> Tuple[npty.NDArray]:
        """Use SciPy's ``find_peaks`` to find the local maxima.

        Unkown arguments are passed to ``find_peaks``.

        If the signal is complex, the absolute value is taken.

        If ``include_edges`` is True, the edges are considered among the
        possible maxima.

        :returns: Coordinate and value of the peaks.
        :rtype: Tuple of NumPy arrays
        """
        y = abs(self.y) if self.is_complex() else self.y
        return self._local_extrema(
            y, *args, include_edges=include_edges, **kwargs
        )

    def local_minima(
        self, *args, include_edges: bool = True, **kwargs
    ) -> Tuple[npty.NDArray]:
        """Use SciPy's ``find_peaks`` to find the local minima.

        Unkown arguments are passed to ``find_peaks``.

        If the signal is complex, the absolute value is taken.

        If ``include_edges`` is True, the edges are considered among the
        possible minima.

        :returns: Coordinate and value of the minima.
        :rtype: Tuple of NumPy arrays
        """
        y = -abs(self.y) if self.is_complex() else -self.y
        return self._local_extrema(
            y, *args, include_edges=include_edges, **kwargs
        )

    def _make_spline(self, *args, k=3, s=0, **kwargs):
        """Private function to make spline representation of the data.

        This function is not meant to be called directly.

        ``k`` is the degree of the spline fit. It is recommended to use cubic
        splines. Even values of ``k`` should be avoided especially with small ``s``
        values. 1 <= k <= 5

        Unknown arguments are pass to ``scipy.interpolate.splrep``.

        :param k: Order of the spline representation.
        :type k:  int
        :param s: Smoothing of the spline.
        :type s:  float

        """
        if len(self) < k:
            raise ValueError(
                f"Too few points to compute a spline of order {k}"
            )

        if self.is_masked():
            raise RuntimeError("Splines with masked data are not supported.")

        self.spline_real = interpolate.splrep(
            self.x, self.y.real, k=k, s=s, *args, **kwargs
        )

        if self.is_complex():
            self.spline_imag = interpolate.splrep(
                self.x, self.y.imag, k=k, s=s, *args, **kwargs
            )

        self.invalid_spline = False

    def evaluate_with_spline(self, x, ext=2):
        """Evaluate the spline on the points ``x``.

        Values outside the interval are extrapolated if ``ext=0``, set to 0 if
        ``ext=1``, raise a ``ValueError`` if ``ext=2``, or if ``ext=3``, return
        the boundary value.

        This method is meant to be used only if you want to use a different ext
        for a specific call, otherwise, just use __call__.

        :param x: Array of x where to evaluate the series or single x.
        :type x: 1D NumPy array of float

        :param ext: How to deal values outside the bounaries. Values outside the
                    interval are extrapolated if ``ext=0``, set to 0 if
                    ``ext=1``, raise a ValueError if ``ext=2``, or if ``ext=3``,
                    return the boundary value.
        :type ext: int

        :returns: Values of the series evaluated on the input x.
        :rtype:   1D NumPy array or float

        """
        if self.invalid_spline:
            self._make_spline()

        y_real = interpolate.splev(x, self.spline_real, ext=ext)
        if self.is_complex():
            y_imag = interpolate.splev(x, self.spline_imag, ext=ext)
            ret = y_real + 1j * y_imag
        else:
            ret = y_real

        # When this method is called with a scalar input, at this point, ret
        # would be a 0d NumPy scalar array. What's that? - you may ask. I have
        # no idea, but the user is expecting a scalar as output. Hence, we cast
        # the 0d array into at "at_least_1d" array, then we can see its length
        # and act consequently.
        ret = np.atleast_1d(ret)
        return ret if len(ret) > 1 else ret[0]

    def __call__(self, x):
        """Evaluate the spline on the points x. If the value is outside the
        range, a ValueError will be raised.
        """
        # We call the spline only if we need to.

        # TODO (REFACTORING): This is not a Pythonic way to write this function.
        #
        # The main problem is that it is is not vectorized. It is also not
        # efficient, we are going over the array a lot of times, re-checking the
        # same elements over and over. Also, we are not considering the floating
        # point arithmetic, we should allow for some tolerance.

        # First we consider the scalar case
        if not hasattr(x, "__len__"):
            if x in self.x:
                return self.y[np.searchsorted(self.x, x)]
            return self.evaluate_with_spline(x, ext=2)

        ret = np.zeros(len(x), dtype=type(self.y[0]))
        # Hash-maps are more efficient than searching every time through the
        # array, but there is some overhead cost in defining the dictionary.
        # Experiments show that it is sill more performant.
        dic_data = dict(zip(self.x, self.y))
        for index, elem in enumerate(x):
            if elem in self.x:
                ret[index] = dic_data[elem]
            else:
                ret[index] = self.evaluate_with_spline(elem, ext=2)

        return ret

    def copy(self):
        """Return a deep copy.

        :returns:  Deep copy of the series.
        :rtype:    :py:class:`~.BaseSeries` or derived class
        """
        # The following is more complicated copy constructor that is designed
        # to copy also the spline information without re-computing it.
        # This can speed up some computations.
        copied = type(self).__new__(self.__class__)
        # We don't use the setters

        # skipcq PYL-W0212
        copied.__data_x = self.__data_x.copy()
        # skipcq PYL-W0212
        copied.__data_y = self.__data_y.copy()
        if not self.invalid_spline:
            # splines are tuples, with a direct call to the function
            # tuple() we make a deep copy
            copied.spline_real = tuple(self.spline_real)
            if self.is_complex():
                copied.spline_imag = tuple(self.spline_imag)
            copied.invalid_spline = False
        copied.invalid_spline = True
        return copied

    def resampled(self, new_x, ext=2, piecewise_constant=False):
        """Return a new series resampled from this to new_x.

        You can specify the details of the spline with the method make_spline.

        If you want to resample without using the spline, and you want a nearest
        neighbor resampling, pass the keyword ``piecewise_constant=True``.
        This may be a good choice for data with large discontinuities, where the
        splines are ineffective.

        :param new_x: New independent variable.
        :type new_x:  1D NumPy array or list of float
        :param ext: How to handle points outside the data interval.
        :type ext: 0 for extrapolation, 1 for returning zero, 2 for ``ValueError``,
                   3 for extending the boundary
        :param piecewise_constant: Do not use splines, use the nearest neighbors.
        :type piecewise_constant: bool
        :returns: Resampled series.
        :rtype:   :py:class:`~.BaseSeries` or derived class

        """
        # If x is the same, there's no need to resample
        if len(self.x) == len(new_x) and np.allclose(
            self.x, new_x, atol=1e-14
        ):
            return self.copy()

        # Unfortunately there is no nearest neighor resampling in SciPy's splines.
        # Hence, we use directly the method interp1d.
        if piecewise_constant:
            interp_function = interpolate.interp1d(
                self.x, self.y, kind="nearest", assume_sorted=True
            )
            new_y = interp_function(new_x)
        else:
            new_y = self.evaluate_with_spline(new_x, ext=ext)

        return type(self)(new_x, new_y)

    def resample(self, new_x, ext=2, piecewise_constant=False):
        """Resample the series to new independent variable new_x.

        If you want to resample without using the spline, and you want a nearest
        neighbor resampling, pass the keyword ``piecewise_constant=True``.
        This may be a good choice for data with large discontinuities, where the
        splines are ineffective.

        :param new_x: New independent variable.
        :type new_x:  1D NumPy array or list of float
        :param ext: How to handle points outside the interval.
        :type ext: 0 for extrapolation, 1 for returning zero, 2 for ValueError,
                   3 for extending the boundary
        :param piecewise_constant: Do not use splines, use the nearest neighbors.
        :type piecewise_constant: bool

        """
        self._apply_to_self(
            self.resampled,
            new_x,
            ext=ext,
            piecewise_constant=piecewise_constant,
        )

    def _apply_binary(self, other, function, *args, **kwargs):
        """This is an abstract function that is used to implement mathematical
        operations with other series (if they have the same x) or
        scalars.

        :py:meth:`~._apply_binary` takes another object that can be of the same
        type or a scalar, and applies ``function(self.y, other.y)``, performing type
        checking.

        :param other: Other object.
        :type other: :py:class:`~.BaseSeries` or derived class or float
        :param function: Dyadic function (function that takes two arguments).
        :type function: callable

        :returns:  Return value of ``function`` when called with self and other.
        :rtype:   :py:class:`~.BaseSeries` or derived class (typically)

        """
        # If the other object is of the same type
        if isinstance(other, type(self)):
            if (len(self.x) != len(other.x)) or (
                not np.allclose(other.x, self.x, atol=1e-14)
            ):
                raise ValueError("The objects do not have the same x!")
            return type(self)(
                self.x,
                function(self.y, other.y, *args, **kwargs),
                True,
            )
        # If it is a number
        if isinstance(other, (int, float, complex)):
            return type(self)(
                self.x, function(self.y, other, *args, **kwargs), True
            )

        # If it is a Tensor of type(self), we have to return a Tensor
        if isinstance(other, Tensor) and type(self) is other.type:
            # We keep this at the high level
            return type(other).from_shape_and_flat_data(
                other.shape,
                [
                    function(ot, self, *args, **kwargs)
                    for ot in other.flat_data
                ],
            )

        # If we are here, it is because we cannot add the two objects
        raise TypeError("I don't know how to combine these objects")

    def __eq__(self, other):
        """Check for equality up to numerical precision."""
        if isinstance(other, type(self)):
            return np.allclose(self.x, other.x, atol=1e-14) and np.ma.allclose(
                self.y, other.y, atol=1e-14
            )
        return False

    # From Python's docs: In order to conform to the object model, classes that
    # define their own equality method should also define their own hash method,
    # or be unhashable.

    # We consider series unhashable. We could make them hashable by taking
    # the combined hash of all the points, but this would not be useful.
    __hash__ = None

    def _apply_to_self(self, f, *args, **kwargs):
        """Apply the method ``f`` to ``self``, modifying ``self``.

        This is used to transform the commands from returning an object to
        modifying ``self``. The function ``f`` has to return a new copy of the
        object (not a reference).

        """
        ret = f(*args, **kwargs)
        # We avoid the setters to avoid checking for consistency because this
        # was already done
        self.__data_x, self.__data_y = ret.x, ret.y
        # We have to recompute the splines
        self.invalid_spline = True

    def save(self, file_name, *args, **kwargs):
        """Saves into simple ASCII format with 2 columns ``(x, y)``
        for real valued data and 3 columns ``(x, Re(y), Im(y))``
        for complex valued data.

        Unknown arguments are passed to ``NumPy.savetxt``.

        :param file_name: Path (with extension) of the output file.
        :type file_name: str

        """
        if self.is_masked():
            warnings.warn(
                "Discarding mask information.",
                RuntimeWarning,
            )

        if self.is_complex():
            np.savetxt(
                file_name,
                np.transpose(
                    (self.x, self.y.real, self.y.imag),
                    *args,
                    **kwargs,
                ),
            )
        else:
            np.savetxt(
                file_name,
                np.transpose((self.x, self.y), *args, **kwargs),
            )

    def nans_removed(self):
        """Filter out nans/infinite values.
        Return a new series with finite values only.

        :returns: A new series with only finite values.
        :rtype: :py:class:`~.BaseSeries` or derived class
        """
        mask = np.isfinite(self.y)
        return type(self)(self.x[mask], self.y[mask], True)

    def nans_remove(self):
        """Filter out nans/infinite values."""
        self._apply_to_self(self.nans_removed)

    def mask_removed(self):
        """Remove masked value.

        Return a new series with valid values only.

        :returns: A new series with only valid values.
        :rtype: :py:class:`~.BaseSeries` or derived class
        """
        if self.is_masked():
            mask = np.invert(self.mask)
            return type(self)(
                np.ma.compressed(self.x[mask]),
                np.ma.compressed(self.y[mask]),
                True,
            )

        # We can copy the spline
        return self.copy()

    def mask_remove(self):
        """Remove masked values."""
        self._apply_to_self(self.mask_removed)

    def mask_applied(self, mask, ignore_existing=False):
        """Return a new series with given mask applied to the data.

        If a previous mask already exists, the new mask will be added on top,
        unless ``ignore_existing`` is True.

        :param mask: Array of booleans that identify where the data is invalid.
                     This can be obtained with the method :py:meth:`~.mask`.
        :type mask: 1D NumPy array

        :param ignore_existing: If True, overwrite any previously existing mask.
        :type ignore_existing: bool

        :returns: New series with mask applied.
        :rtype: :py:class:`~.BaseSeries`

        """
        if self.is_masked() and not ignore_existing:
            mask = np.ma.mask_or(mask, self.mask)

        return type(self)(self.x, np.ma.MaskedArray(self.y, mask=mask), True)

    def mask_apply(self, mask, ignore_existing=False):
        """Apply given mask.

        If a previous mask already exists, the new mask will be added on top,
        unless ``ignore_existing`` is True.

        :param mask: Array of booleans that identify where the data is invalid.
                     This can be obtained with the method :py:meth:`~.mask`.
        :type mask: 1D NumPy array

        :param ignore_existing: If True, overwrite any previously existing mask.
        :type ignore_existing: bool
        """
        self._apply_to_self(
            self.mask_applied, mask, ignore_existing=ignore_existing
        )

    def integrated(self, dx=None):
        """Return a series that is the integral computed with method of
        the rectangles.

        The spacing ``dx`` can be optionally provided. If provided, it will be
        used (increasing performance), otherwise it will be computed internally.

        :param dx: Delta x in the independent variable. If None it will be
                   computed internally.
        :type dx: float or None
        :returns:  New series with the cumulative integral.
        :rtype:    :py:class:`~.BaseSeries` or derived class

        """
        if self.is_masked():
            raise RuntimeError(
                "Integration with masked data is not supported."
            )

        # We pass self.x only if dx was not provided
        passing_x = self.x if dx is None else None
        return type(self)(
            self.x,
            integrate.cumtrapz(self.y, x=passing_x, dx=dx, initial=0),
            True,
        )

    def integrate(self, dx=None):
        """Integrate series with method of the rectangles.

        The spacing ``dx`` can be optionally provided. If provided, it will be
        used (increasing performance), otherwise it will be computed internally.

        """
        self._apply_to_self(self.integrated, dx=dx)

    def spline_differentiated(self, order=1):
        """Return a series that is the derivative of the current one using
        the spline representation.

        The optional parameter ``order`` specifies the order of the derivative.

        .. warning::

            The values at the boundary are typically not accurate.

        :param order: Order of derivative (e.g. 2 = second derivative).
        :type order: int

        :returns:  New series with derivative
        :rtype:    :py:class:`~.BaseSeries` or derived class

        """
        if (order > 3) or (order < 0):
            raise ValueError(f"Cannot compute differential of order {order}")

        if self.invalid_spline:
            self._make_spline()

        if self.is_complex():
            ret_value = interpolate.splev(
                self.x, self.spline_real, der=order
            ) + 1j * interpolate.splev(self.x, self.spline_imag, der=order)
        else:
            ret_value = interpolate.splev(self.x, self.spline_real, der=order)

        return type(self)(self.x, ret_value, True)

    def spline_differentiate(self, order=1):
        """Differentiate the series using the spline representation.

        The optional parameter ``order`` specifies the order of the derivative.

        .. warning::

            The values at the boundary are typically not accurate.

        :param order: Order of derivative (e.g. 2 = second derivative).
        :type order: int

        """
        self._apply_to_self(self.spline_differentiated, order)

    def differentiated(self, order=1):
        """Return a series that is the numerical order-differentiation of
        the present series.

        The optional parameter ``order`` specifies the order of the derivative.

        The derivative is calulated as centered differencing in the interior
        and one-sided derivatives at the boundaries. Higher orders are computed
        applying the same rule recursively.

        :param order: Order of derivative (e.g. 2 = second derivative).
        :type order: int

        :returns:  New series with derivative.
        :rtype:    :py:class:`~.BaseSeries` or derived class

        """
        if self.is_masked():
            raise RuntimeError(
                "Differentiation with masked data is not supported."
            )

        ret_value = self.y
        for _num_deriv in range(order):
            ret_value = np.gradient(ret_value, self.x, edge_order=2)
        return type(self)(self.x, ret_value, True)

    def differentiate(self, order=1):
        """Differentiate with the numerical order-differentiation.

        The optional parameter ``order`` specifies the order of the derivative.

        The derivative is calulated as centered differencing in the interior
        and one-sided derivatives at the boundaries. Higher orders are computed
        applying the same rule recursively.

        :param order: Order of derivative (e.g. 2 = second derivative).
        :type order: int

        """
        self._apply_to_self(self.differentiated, order)

    def savgol_smoothed(self, window_size, order=3):
        """Return a smoothed series with a Savitzky-Golay filter with
        window of size ``window_size`` and order ``order``.

        This is just like a regular "Moving average" filter, but instead of
        just calculating the average, a polynomial (usually 2nd or 4th order)
        fit is made for every point, and only the "middle" point is chosen.
        Since 2nd (or 4th) order information is concerned at every point, the
        bias introduced in "moving average" approach at local maxima or minima,
        is circumvented.

        :param window_size: Number of points of the smoothing window (needs to
                            be odd).
        :type window_size: int
        :param order: Order of the filter.
        :type order: int

        :returns:  New smoothed series.
        :rtype:    :py:class:`~.BaseSeries` or derived class

        """
        if self.is_masked():
            raise RuntimeError("Smoothing with masked data is not supported.")

        if self.is_complex():
            return type(self)(
                self.x,
                signal.savgol_filter(self.y.imag, window_size, order)
                + 1j * signal.savgol_filter(self.y.real, window_size, order),
                True,
            )

        return type(self)(
            self.x,
            signal.savgol_filter(self.y, window_size, order),
            True,
        )

    def savgol_smooth(self, window_size, order=3):
        """Smooth the series with a Savitzky-Golay filter with window of
        size ``window_size`` and order ``order``.

        This is just like a regular "Moving average" filter, but instead of
        just calculating the average, a polynomial (usually 2nd or 4th order)
        fit is made for every point, and only the "middle" point is chosen.
        Since 2nd (or 4th) order information is concerned at every point, the
        bias introduced in "moving average" approach at local maxima or minima,
        is circumvented.

        :param window_size: Number of points of the smoothing window (needs to
                            be odd).
        :type window_size: int
        :param order: Order of the filter.
        :type order: int

        """
        self._apply_to_self(self.savgol_smoothed, window_size, order)

    def cropped(self, init=None, end=None):
        """Return a series with data removed outside the interval ``[init, end]``. If
        ``init`` or ``end`` are not specified or None, it does not remove
        anything from this side.

        :param init: Data with ``x <= init`` will be removed.
        :type init: float or None
        :param end: Data with ``x >= end`` will be removed.
        :type end: float or None

        :returns:  Series with enforced minimum and maximum
        :rtype:    :py:class:`~.BaseSeries` or derived class

        """
        x = self.x
        y = self.y
        if init is not None:
            m = x >= init
            x = x[m]
            y = y[m]
        if end is not None:
            m = x <= end
            x = x[m]
            y = y[m]
        return type(self)(x, y, True)

    def crop(self, init=None, end=None):
        """Remove data outside the the interval ``[init, end]``. If
        ``init`` or ``end`` are not specified or None, it does not remove
        anything from this side.

        :param init: Data with ``x <= init`` will be removed.
        :type init: float or None
        :param end: Data with ``x >= end`` will be removed.
        :type end: float or None

        """
        self._apply_to_self(self.cropped, init, end)

    # Define aliases
    clip = crop
    clipped = cropped

    def _apply_unary(self, function, *args, **kwargs):
        """Apply a unary function to the data.

        :param function: Function to apply to the series.
        :type function: callable

        :return: New series with function applied to the data.
        :rtype: :py:class:`~.BaseSeries` or derived class

        """
        return type(self)(self.x, function(self.y, *args, **kwargs), True)

    def _apply_reduction(self, reduction, *args, **kwargs):
        """Apply a reduction to the data.

        :param function: Function to apply to the series.
        :type function: callable

        :return: Reduction applied to the data
        :rtype: float

        """
        return reduction(self.y, *args, **kwargs)


def sample_common(series, resample=False, piecewise_constant=False):
    """Take a list of series and return new ones so that they are all defined on the
    same points.

    If ``resample`` is False (default), take as input a list of series and
    return a new list with the same series but only defined on those points that
    are common to all the lists. If ``resample`` is True, instead of removing
    points, find the common interval of definition, and resample all the series
    on that internal. The number of sample points is the minimum over all
    series. Additionally, if ``piecewise_constant=True``, the approximant used
    for resampling is a piecewise constant function, splines are not used,
    instead, the nearest neighbors are used. Use this when you have series with
    discontinuities.

    :param series: The series to resample or redefine on the common points
    :type series:  list of :py:class:`~.Series`
    :param resample: Whether to resample the series, or just find the common
                     points.
    :type resample: bool
    :param piecewise_constant: Whether to use the nearest neighbor resampling
                               method instead of splines.
                               If ``piecewise_constant=True``, the approximant used
                               for resampling is a piecewise constant function.
    :type piecewise_constant: bool
    :returns:  Resampled series so that they are all defined in
               the same interval.
    :rtype:    list of :py:class:`~.Series`

    """
    # In many cases there is no real need for resampling because the array
    # already have the desired shape. It is worth checking if we need to
    # resample. If the series are regularly sampled, it is easy to check
    # if the are the same. We also need to check that they are regularly
    # sampled, to do this, we check that the first is regularly sampled,
    # and that all the other ones have the same x.
    s1, *s_others = series
    if s1.is_regularly_sampled():
        for s in s_others:
            if not (len(s) == len(s1)):
                break
            if not np.allclose(s1.x, s.x, atol=1e-14):
                break
            # This is an else to the for loop
        else:
            # We have to copy, otherwise one can accidentally modify input data
            return [ss.copy() for ss in series]

    if resample:
        # Find the series with max xmin
        s_xmin = max(series, key=lambda x: x.xmin)
        # Find the series with min xmax
        s_xmax = min(series, key=lambda x: x.xmax)
        # Find the series with min number of points
        s_ns = min(series, key=len)
        x = np.linspace(s_xmin.xmin, s_xmax.xmax, len(s_ns))
        return [
            s.resampled(x, piecewise_constant=piecewise_constant)
            for s in series
        ]

    def float_intersection(array_1, array_2):
        """Here we find the intersection between the two arrays also
        considering the floating points.
        """
        # https://stackoverflow.com/a/32516182
        return array_2[
            np.isclose(array_1[:, None], array_2, atol=1e-14).any(0)
        ]

    # Here we find the common intersection between all the x, starting with
    # the first one
    x = s1.x
    for s in s_others:
        x = float_intersection(x, s.x)
        if len(x) == 0:
            raise ValueError("Series do not have any point in common")

    return [s.resampled(x) for s in series]



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# Based on code originally developed by Wolfgang Kastaun. This file may contain
# algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/simdir.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""This module provides easy access to Cactus data files.

A simulation directory is represented by an instance of the :py:class:`~.SimDir`
class, which provides access to all supported data types.

This is the main entry point into ``kuibit``. When a :py:class:`~.SimDir` is
initialized, the simulation directory is scanned and all the data is organized.
:py:class:`~.SimDir` objects have attributes that are interfaces to the data:
each attribute is a dictionary-like object that indexes the relevant data in
some way. For example, :py:meth:`~.timeseries` contains all the time series in
the output, indexed by the type of reduction that produced them (for example,
``norm2``, ``max``, ...).

In case of uncertainty, it is always possible to print :py:class:`~.SimDir`,
or any of its attributes, to obtain a message with the available content of
such attribute.

The function :py:func:`~.load_SimDir` can be used to load a :py:class:`~.SimDir`
saved with the method :py:meth:`~.save`.

"""
import os
import pickle

from kuibit import (
    cactus_grid_functions,
    cactus_horizons,
    cactus_multipoles,
    cactus_scalars,
    cactus_timers,
    cactus_twopunctures,
    cactus_waves,
)


def load_SimDir(path):
    """Load file produced with :py:meth:`~.SimDir.save`.

    Pickles have to be regenerated if the version of ``kuibit`` changes.

    :param path: Pickle file as produced by :py:meth:`~.SimDir.save`.
    :type path: str

    :returns: SimDir
    :rtype: :py:class:`~.SimDir`

    """
    with open(path, "rb") as file_:
        sim = pickle.load(file_)

    if not isinstance(sim, SimDir):
        raise RuntimeError(f"File {path} does not contain a SimDir")

    return sim


class SimDir:
    """This class represents a Cactus simulation directory.

    :py:class:`~.SimDir` can be used as a context manager. For instance:

    .. code-block

       with SimDir(sim_path) as sim:
            print(sim)

    By itself, this is not very useful. It becomes more useful in conjunction
    with using pickles. Pickles are used to save the work done and resume it
    later. Since ``kuibit`` does a lot of lazy-loading, it can be useful to save
    the operations performed to disk and restart from them. For example, when
    initializing a :py:class:`~.SimDir`, the files have to be scanned and
    organized. It is pointless to this all the times if the simulation has not
    changed. For this, we use pickles.

    .. code-block

       with SimDir(sim_path, pickle_file="sim.pickle") as sim:
            print(sim)

    What happens here is that if ``pickle_file`` exists, it will be loaded. If
    it does not exist, it will be created. Using context managers here is useful
    because it automatically saves all the progress done. Alternatively, one
    has to call the :py:meth:`~.save` method manually.

    .. warning::

       When using pickles, it is important to make sure that the underlying data
       does not change, as the new/changed data will be not considered. To
       refresh :py:class:`~.SimDir`, you can always use the :py:meth:`~.rescan`
       method. Pickles have to be regenerated from scratch if the version of
       ``kuibit`` changes.

    Data is searched recursively in all subfolders. No particular folder
    structure (e.g. ``simfactory`` style) is assumed. The following attributes
    allow access to the supported data types:

    :ivar path:           Top level path of simulation directory.
    :ivar dirs:           All directories in which data is searched.
    :ivar logfiles:       The locations of all log files (.out).
    :ivar errfiles:       The location of all error log files (.err).
    :ivar ts:             Scalar data of various type, see
                          :py:class:`~.ScalarsDir`
    :ivar gf:              Access to grid function data, see
                          :py:class:`~.GridFunctionsDir`.
    :ivar gws:            GW signal from the Weyl scalar multipole
                          decomposition, see
                          :py:class:`~.GravitationalWavesDir`.
    :ivar emws:           EM signal from the Weyl scalar multipole
                          decomposition, see
                          :py:class:`~.ElectromagneticWavesDir`.
    :ivar horizons:       Apparent horizon information, see
                          :py:class:`~.HorizonsDir`.
    :ivar timers:         Timer information, see
                          :py:class:`~.TimersDir`.
    :ivar twopunctures:   Metadata information from TwoPunctures.
                          :py:class:`~.TwoPuncturesDir`.
    :ivar multipoles:     Multipole components, see
                          :py:class:`~.CactusMultipoleDir`.

    """

    @staticmethod
    def _sanitize_path(path):
        # Make sure to have complete paths with respect to the current folder
        abs_path = os.path.abspath(os.path.expanduser(path))
        if not os.path.isdir(abs_path):
            raise RuntimeError(f"Folder does not exist: {path}")

        return abs_path

    def _scan_folders(self, max_depth):
        """Scan all the folders in self.path up to depth ``max_depth``
        and categorize all the files.

        :param max_depth: Maximum recursion depth to scan.
        :type max_depth: int
        """

        def listdir_process_symlinks(path):
            """Return a list of files in path. If self.ignore_symlinks, exclude the
            symlinks, otherwise keep them around.

            """
            dir_content = [os.path.join(path, p) for p in os.listdir(path)]
            if not self.ignore_symlinks:
                return dir_content
            return [p for p in dir_content if not os.path.islink(p)]

        def filter_ext(files, ext):
            """Return a list from the input list of file that
            has file extension ext."""
            return [f for f in files if os.path.splitext(f)[1] == ext]

        def walk_rec(path, level=0):
            """Walk_rec is a recursive function that steps down all the
            subdirectories (except the ones with name defined in
            self.ignored_dirs) up to max_depth and add to self.allfiles the
            files found in the directories.

            """
            if level >= max_depth:
                return

            self.dirs.append(path)

            all_files_in_path = listdir_process_symlinks(path)

            files_in_path = list(filter(os.path.isfile, all_files_in_path))
            self.allfiles += files_in_path

            directories_in_path = list(
                filter(os.path.isdir, all_files_in_path)
            )

            # We ignore the ones in self.ignored_dirs
            directories_to_scan = [
                p
                for p in directories_in_path
                if (os.path.basename(p) not in self.ignored_dirs)
            ]

            # Apply walk_rec to all the subdirectory, but with level increased
            for p in directories_to_scan:
                walk_rec(p, level + 1)

        walk_rec(self.path)

        self.logfiles = filter_ext(self.allfiles, ".out")
        self.errfiles = filter_ext(self.allfiles, ".err")
        self.parfiles = filter_ext(self.allfiles, ".par")

        # Sort by time
        self.parfiles.sort(key=os.path.getmtime)
        self.logfiles.sort(key=os.path.getmtime)
        self.errfiles.sort(key=os.path.getmtime)

        simfac = os.path.join(self.path, "SIMFACTORY", "par")

        # Simfactory has a folder SIMFACTORY with a subdirectory for par files
        # Even if SIMFACTORY is excluded, we should include that par file
        if os.path.isdir(simfac):
            mainpar = filter_ext(listdir_process_symlinks(simfac), ".par")
            self.parfiles = mainpar + self.parfiles

        self.has_parfile = bool(self.parfiles)

    def __init__(
        self,
        path,
        max_depth=8,
        ignored_dirs=None,
        ignore_symlinks=True,
        pickle_file=None,
    ):
        """Constructor.

        :param path:      Path to output of the simulation.
        :type path:       str
        :param max_depth: Maximum recursion depth for subfolders.
        :type max_depth:  int
        :param ignored_dirs: Names of folders to ignore (e.g. SIMFACTORY).
        :type ignored_dirs:  set
        :param ignore_symlink: If True, do not consider symlinks.
        :type ignore_symlink: bool
        :param pickle_file: If ``pickle_file`` is not None, do not scan the
                            folders and load the pickle file. All the other
                            parameters are ignored.
        :type pickle_file: bool

        Parfiles (``*.par``) will be searched in all data directories and the
        top-level SIMFACTORY/par folder, if it exists. The parfile in the latter
        folder, if available, or else the oldest parfile in any of the data
        directories, will be used to extract the simulation parameters. Logfiles
        (``*.out``) and errorfiles (``*.err``) will be searched for in all data
        directories.

        """
        if ignored_dirs is None:
            ignored_dirs = {"SIMFACTORY", "report", "movies", "tmp", "temp"}

        # We update self.path in _sanitize_path to make sure it is an absolute
        # path
        self.path = self._sanitize_path(str(path))

        self.max_depth = int(max_depth)
        self.ignored_dirs = ignored_dirs
        self.ignore_symlinks = ignore_symlinks

        self.dirs = []
        self.parfiles = []
        self.logfiles = []
        self.errfiles = []
        self.allfiles = []
        self.has_parfile = False
        self.__timeseries = None
        self.__multipoles = None
        self.__gravitationalwaves = None
        self.__electromagneticwaves = None
        self.__gridfunctions = None
        self.__horizons = None
        self.__timers = None
        self.__twopunctures = None

        if (pickle_file is None) or (not os.path.exists(pickle_file)):
            self._populate()
        else:
            sim = load_SimDir(pickle_file)
            # Overwrite all the local variables
            self.__dict__ = sim.__dict__

        # We set this later, so that if it was read from the pickle, we override
        # it in such a way that we have consistency.
        self.pickle_file = pickle_file

    def _populate(self):
        """Scan the folders and populate basic attributes."""

        self._scan_folders(self.max_depth)

        self.__timeseries = None
        self.__multipoles = None
        self.__gravitationalwaves = None
        self.__electromagneticwaves = None
        self.__gridfunctions = None
        self.__horizons = None
        self.__timers = None
        self.__twopunctures = None

    def rescan(self):
        """Reset the SimDir and rescan all the files."""
        self._populate()

    @property
    def timeseries(self):
        """Return all the available timeseries in the data.

        :returns: Interface to all the timeseries in the directory.
        :rtype: :py:class:`~.ScalarsDir`
        """
        if self.__timeseries is None:
            self.__timeseries = cactus_scalars.ScalarsDir(self)
        return self.__timeseries

    ts = timeseries

    @property
    def multipoles(self):
        """Return all the available multipole data.

        :returns: Interface to all the multipole data in the directory.
        :rtype: :py:class:`~.MultipolesDir`
        """
        if self.__multipoles is None:
            self.__multipoles = cactus_multipoles.MultipolesDir(self)
        return self.__multipoles

    @property
    def gravitationalwaves(self):
        """Return all the available ``Psi4`` data.

        :returns: Interface to all the ``Psi4`` data in the directory.
        :rtype: :py:class:`~.GravitationalWavesDir`
        """
        if self.__gravitationalwaves is None:
            self.__gravitationalwaves = cactus_waves.GravitationalWavesDir(
                self
            )
        return self.__gravitationalwaves

    gws = gravitationalwaves

    @property
    def electromagneticwaves(self):
        """Return all the available ``Phi2`` data.

        :returns: Interface to all the ``Phi2`` data in the directory.
        :rtype: :py:class:`~.ElectromagneticWavesDir`
        """
        if self.__electromagneticwaves is None:
            self.__electromagneticwaves = cactus_waves.ElectromagneticWavesDir(
                self
            )
        return self.__electromagneticwaves

    emws = electromagneticwaves

    @property
    def gridfunctions(self):
        """Return all the available grid data.

        :returns: Interface to all the grid data in the directory.
        :rtype: :py:class:`~.GridFunctionsDir`
        """
        if self.__gridfunctions is None:
            self.__gridfunctions = cactus_grid_functions.GridFunctionsDir(self)
        return self.__gridfunctions

    gf = gridfunctions

    @property
    def horizons(self):
        """Return all the available horizon data.

        :returns: Interface to all the horizon data in the directory.
        :rtype: :py:class:`~.HorizonsDir`
        """
        if self.__horizons is None:
            self.__horizons = cactus_horizons.HorizonsDir(self)
        return self.__horizons

    @property
    def timers(self) -> cactus_timers.TimersDir:
        """Return all the available timertree data.

        :returns: Interface to all the timertree data in the directory.
        :rtype: :py:class:`~.TimertreeDir`
        """
        if self.__timers is None:
            self.__timers = cactus_timers.TimersDir(self)
        return self.__timers

    @property
    def twopunctures(self):
        """Return the metadata for TwoPunctures.

        :returns: Interface to the metadata in TwoPunctures.
        :rtype: :py:class:`~.TwoPuncturesDir`
        """
        if self.__twopunctures is None:
            self.__twopunctures = cactus_twopunctures.TwoPuncturesDir(self)
        return self.__twopunctures

    def __str__(self):
        header = f"Indexed {len(self.allfiles)} files"
        header += f" and {len(self.dirs)} subdirectories\n"

        ts_ret = f"{self.ts}"
        mp_ret = f"{self.multipoles}"
        gf_ret = f"{self.gf}"

        if len(self.gravitationalwaves) > 0:
            gw_ret = "Available gravitational wave data"
        else:
            gw_ret = ""

        if len(self.electromagneticwaves) > 0:
            em_ret = "Available electromagnetic wave data"
        else:
            em_ret = ""

        hor_ret = f"{self.horizons}"

        tim_ret = f"\n{self.timers}"

        return (
            header
            + ts_ret
            + mp_ret
            + gw_ret
            + em_ret
            + gf_ret
            + hor_ret
            + tim_ret
        )

    def __enter__(self):
        """This is classed when the object is used as a context manager."""
        # All the work in done in __init__
        return self

    def __exit__(self, _1, _2, _3):
        """Save the SimDir to disk as pickle.

        This is called when the object is used as a context manager.

        """
        if self.pickle_file is not None:
            self.save(self.pickle_file)

    def save(self, path):
        """Save this object as a pickle.

        The object can be loaded with the function :py:func:`~.load_SimDir`.

        :param path: Path where to save the file.
        :type path: str

        """
        with open(path, "wb") as file_:
            pickle.dump(self, file_, protocol=pickle.HIGHEST_PROTOCOL)



#!/usr/bin/env python3

# Copyright (C) 2022-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.hor_utils` module provides functions to perform common
operations using horizon(s).

The functions provided are:

- :py:func:`~.compute_center_of_mass`: compute the Newtonian center of mass for the
  given two horizons.
- :py:func:`~.compute_separation_vector`: compute the Newtonian separation vector
  between the centroids of the given two horizons.
- :py:func:`~.compute_separation`: compute the Newtonian separation between the centroids
  of the given two horizons.
- :py:func:`~.compute_angular_velocity_vector`: compute the Newtonian angular velocity vector
  for the given two horizons.

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

from kuibit.series import sample_common
from kuibit.tensor import Vector
from kuibit.timeseries import TimeSeries

if TYPE_CHECKING:  # pragma: no cover
    from kuibit.cactus_horizon import OneHorizon


def _two_centroids_as_Vectors(
    horizon1: OneHorizon, horizon2: OneHorizon, resample: bool = True
) -> Tuple[Vector[TimeSeries], Vector[TimeSeries]]:
    """Process horizon centroids.

    The information from the apparent horizons is used (contained in the
    BHDiagnostics files).

    Here we transform the position of the centroids into :py:class:`~.Vector`
    because they are easier to work with for vector operations.

    :param horizon1: First horizon. It has to have AH information.
    :type horizon1: :py:class:`~.OneHorizon`
    :param horizon2: Second horizon. It has to have AH information.
    :type horizon2: :py:class:`~.OneHorizon`

    :returns: Vectors with the centroid positions
    :rtype: Tuple of :py:class:`~.Vector`

    """

    if not horizon1.ah_available:
        raise RuntimeError(
            "Centroid information not available for first horizon"
        )

    if not horizon2.ah_available:
        raise RuntimeError(
            "Centroid information not available for second horizon"
        )

    # We add sample_common to make sure that everything is defined on the same
    # interval.
    (
        cen1_x,
        cen1_y,
        cen1_z,
        cen2_x,
        cen2_y,
        cen2_z,
    ) = sample_common(
        (
            horizon1.ah.centroid_x,
            horizon1.ah.centroid_y,
            horizon1.ah.centroid_z,
            horizon2.ah.centroid_x,
            horizon2.ah.centroid_y,
            horizon2.ah.centroid_z,
        ),
        resample=resample,
    )
    return Vector([cen1_x, cen1_y, cen1_z]), Vector([cen2_x, cen2_y, cen2_z])


def compute_separation_vector(
    horizon1: OneHorizon, horizon2: OneHorizon, resample: bool = True
) -> Vector[TimeSeries]:
    """Compute the vector coordinate separation between the centroids of two horizons.

    The information from the apparent horizons is used (contained in the
    BHDiagnostics files).

    The separation has sign and we compute it as first horizon - second horizon.


    :param horizon1: First horizon.
    :type horizon1: :py:class:`~.OneHorizon`
    :param horizon2: Second horizon.
    :type horizon2: :py:class:`~.OneHorizon`

    :returns: Coordinate distance vector between the two centroids,
              sampled over where both the horizons are available.
    :rtype: :py:class:`~.Vector` of :py:class:`~.TimeSeries`

    """
    ah1, ah2 = _two_centroids_as_Vectors(horizon1, horizon2, resample)
    return ah1 - ah2


def compute_separation(
    horizon1: OneHorizon, horizon2: OneHorizon, resample: bool = True
):
    """Compute the coordinate separation between the centroids of two horizons.

    The information from the apparent horizons is used (contained in the
    BHDiagnostics files).

    :param horizon1: First horizon.
    :type horizon1: :py:class:`~.OneHorizon`
    :param horizon2: Second horizon.
    :type horizon2: :py:class:`~.OneHorizon`

    :returns: Coordinate distance between the two centroids, sampled over both
              the horizons are available.
    :rtype: :py:class:`~.TimeSeries`

    """
    return compute_separation_vector(horizon1, horizon2, resample).norm()


def compute_center_of_mass(
    horizon1: OneHorizon, horizon2: OneHorizon, resample: bool = True
) -> Vector[TimeSeries]:
    """Compute the Newtonian center of mass between the centroids of two horizons.

    The information from the apparent horizons is used (contained in the
    BHDiagnostics files). The center of mass is computed with the irreducible
    mass (which is directly proportional to the area as computed by
    AHFinderDirect).

    :param horizon1: First horizon.
    :type horizon1: :py:class:`~.OneHorizon`
    :param horizon2: Second horizon.
    :type horizon2: :py:class:`~.OneHorizon`

    :returns: Center of mass vector between the two centroids, sampled over
              where both the horizons are available.
    :rtype: :py:class:`~.Vector` of :py:class:`~.TimeSeries`

    """
    cen1, cen2 = _two_centroids_as_Vectors(horizon1, horizon2, resample)

    # Here we use the fact that the irreducible_mass is proportional to the
    # area. We prefer using the area because it is always available when the
    # centroids are available.

    area1, area2 = sample_common(
        (
            horizon1.ah.area,
            horizon2.ah.area,
        ),
        resample=resample,
    )

    # This is morally "total mass" (irreducible_mass = area / 4pi)
    total_area = area1 + area2

    return area1 * cen1 / total_area + area2 * cen2 / total_area


def compute_angular_velocity_vector(
    horizon1: OneHorizon, horizon2: OneHorizon, resample: bool = True
):
    r"""Compute the angular velocity vector.

    The information from the apparent horizons is used (contained in the
    BHDiagnostics files).

    The angular velocity vector is computed as

    .. math::

      \Omega = \frac{\mathbf{r} \times \mathbf{r}}{r^2}

    :param horizon1: First horizon.
    :type horizon1: :py:class:`~.OneHorizon`
    :param horizon2: Second horizon.
    :type horizon2: :py:class:`~.OneHorizon`

    :returns: Vectors with the centroid positions
    :rtype: Tuple of :py:class:`~.Vector`

    """
    sep = compute_separation_vector(horizon1, horizon2, resample)
    dot_sep = sep.differentiated()
    sep_sq = sep.norm() ** 2

    # TODO (Future): Use Vector.cross

    numerator = Vector(
        [
            sep[1] * dot_sep[2] - sep[2] * dot_sep[1],
            sep[2] * dot_sep[0] - sep[0] * dot_sep[2],
            sep[0] * dot_sep[1] - sep[1] * dot_sep[0],
        ]
    )

    return numerator / sep_sq



#!/usr/bin/env python3

# Copyright (C) 2021-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.masks` module provides functions to work with masked data.

The module contains syntactic sugar to call masked methods in objects.
"""


class _MaskedFunction:
    def __init__(self, name):
        self.__doc__ = f"""Invoke the method ``ma_{name}`` in the given object.
        This is just syntactic sugar.

        :param struct: Object with a method named ``ma_{name}``.
        :type struct: anything

        :returns: Return value of ``ma_{name}``.
        :rtype: Same type as input
        """

        self.__name__ = name

    def __str__(self):
        return f"Masked version of {self.__name__}"

    def __call__(self, struct):
        _name = self.__name__
        if not hasattr(struct, f"ma_{_name}"):
            raise AttributeError(
                f"{type(struct)} does not support masked {_name}"
            )

        return getattr(struct, f"ma_{_name}")()


sqrt = _MaskedFunction("sqrt")
exp = _MaskedFunction("exp")
log = _MaskedFunction("log")
log2 = _MaskedFunction("log2")
log10 = _MaskedFunction("log10")
sin = _MaskedFunction("sin")
cos = _MaskedFunction("cos")
tan = _MaskedFunction("tan")
arcsin = _MaskedFunction("arcsin")
arccos = _MaskedFunction("arccos")
arctan = _MaskedFunction("arctan")
arccosh = _MaskedFunction("arccosh")
arctanh = _MaskedFunction("arctanh")



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
# Copyright (C) 2020 Wolfgang Kastaun
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

r"""This module provides a class Units representing unit systems or unit
conversions.

Units can be used to convert from geometrized units to SI.

For example, assuming a that we are using geometrized units with :math:`G = c =
M = 1`, where :math:`M = 65 M_\odot`, we can defined the object ``CU =
geom_umass(65 * M_SUN_SI)``. CU knows how to convert geometrized quantities to
SI: for instance, to convert the length variable d from geometrized units to SI,
just multiply d times ``CU.length``. Similarly for all the other quantities.

The module provides convenience functions to define geometrized system of units:
:py:func:`~.geom_umass_msun`, :py:func:`~.geom_umass`, and
:py:func:`~.geom_length` can be used to generate :py:class:`~.Units` objects
that know how to convert from/to SI and geometrized units where lengthscale or
mass scales were defined.

The following natural constants are defined (in SI units)

- ``C_SI``          (Vacuum speed of light)
- ``G_SI``          (Gravitational constant)
- ``M_SOL_SI``      (Solar mass)
- ``M_SUN_SI``      (Solar mass)
- ``LIGHTYEAR_SI``  (Lightyear)
- ``MEGAPARSEC_SI`` (Megaparsec)
- ``PARSEC_SI``     (Parsec)
- ``H0_SI``         (Hubble constant [1/s])

"""
from scipy import constants


class Units:
    """Class representing unit conversion. The unit system is specified by
    length, time, and mass units, from which derived units are computed.

    This class can be used to convert units from one system to another.

    For example, define ``CGS = Units(1.0e-2, 1.0, 1.0e-3)``.
    If the length d is in SI, then ``d * CGS.length`` will be in CGS.

    The main use of Units in kuibit is to provide a way to convert from
    geometrized units to physical units (see, :py:func:`~.geom_umass`).
    """

    def __init__(self, ulength, utime, umass):
        """Create a unit system based on length unit ulength, time unit utime,
        and mass unit umass.

        :param ulength: Unit of length with respect to SI.
        :type ulength: float
        :param utime: Unit of time with respect to SI.
        :type utime: float
        :param umass: Unit of mass with respect to SI.
        :type umass: float

        """
        # NOTE: If you add any quantity here, modify the unitconv.rst doc!

        self.length = float(ulength)
        self.time = float(utime)
        self.mass = float(umass)
        self.freq = 1.0 / self.time
        self.velocity = self.length / self.time
        self.accel = self.velocity / self.time
        self.force = self.accel * self.mass
        self.area = self.length * self.length
        self.volume = self.area * self.length
        self.density = self.mass / self.volume
        self.pressure = self.force / self.area
        self.power = self.force * self.velocity
        self.energy = self.force * self.length
        self.energy_density = self.energy / self.volume
        self.angular_moment = self.energy * self.time
        self.moment_inertia = self.mass * self.area

        # TODO (FEATURE): All the electromagnetic quantities are missing.


# Why do we re-define units here when we have scipy.constants? Two reasons:
# 1. We want to stress that the base units are SI
# 2. scipy.constants does not have astronomical constants
#    (we could use astropy.constants, but astropy is too big of a dependency)

# NOTE: If you add any constant here, modify the unitconv.rst doc!

# The following constants are all given in SI units
C_SI = constants.speed_of_light  # Speed of light in vacuum
G_SI = constants.gravitational_constant  # Gravitational constant

M_SOL_SI = 1.988_435e30  # Solar mass
M_SUN_SI = M_SOL_SI  # Alias for solar mass

PARSEC_SI = constants.parsec  # Parsec
MEGAPARSEC_SI = 1e6 * constants.parsec  # Megaparsec
GIGAPARSEC_SI = 1e9 * constants.parsec  # Gigaparsec

LIGHTYEAR_SI = constants.light_year  # Lightyear

H0_SI = 2.192_711_267e-18  # Hubble constant [1/s]


def geom_ulength(ulength):
    """Create a geometric unit system, expressed in SI, based on the given length
    unit expressed in SI units as well.

    :param ulength: Unit of length with respect to SI.
    :type ulength: float
    :rvalue: Geometrized units with length scale set by ``ulength``.
    :rtype: :py:class :`~.Units`

    """
    return Units(ulength, ulength / C_SI, ulength * (C_SI**2) / G_SI)


def geom_umass(umass):
    """Create a geometric unit system, expressed in SI, based on the given mass
    unit expressed in SI units as well.

    :param ulength: Unit of mass with respect to SI.
    :type ulength: float

    :rvalue: Geometrized units with mass scale set by ``umass``.
    :rtype: :py:class:`~.Units`

    """
    return geom_ulength(umass * G_SI / (C_SI**2))


def geom_umass_msun(umass):
    """Create a geometric unit system, expressed in SI, based on the given mass
    unit expressed in solar masses.

    :param ulength: Unit of mass in solar masses.
    :type ulength: float
    :rvalue: Geometrized units with mass scale set by ``umass``.
    :rtype: :py:class:`~.Units`

    """
    return geom_umass(umass * M_SOL_SI)



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# Inspired by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/visualize.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.visualize_matplotlib` module provides methods to plot
``kuibit`` objects with matplotlib and other convenience functions.

Utilities:

- :py:func:`~.setup_matplotlib` adjusts the configuration in matplotlib.
- :py:func:`~.add_text_to_corner` adds a label near an edge or a corner of
  a figure (useful for annotations like time).
- :py:func:`~.save` saves the figure, optionally with tikzplotlib.
- :py:func:`~.save_from_dir_filename_ext` saves the figure and assembles the
  name automatically from the output directory, file name and extension.
- :py:func:`~.set_axis_limits` sets the range on the two axes of a given axis.
  :py:func:`~.set_axis_limits_from_args` does the same but reading the data
  from a given ``args`` (from ``ArgParse``).

Two decorators:

- :py:func:`~.preprocess_plot`. The goal of this is to add support to the
  keyword arguments ``figure`` and ``axis``. If you decorate a function with
  :py:func:`~.preprocess_plot` and the function takes those two arguments, then
  the decorator will automatically check if the arguments were passed and if not
  set the correct default.

- :py:func:`~.preprocess_plot_grid`. This decorator takes some form of grid data
  and returns a NumPy array (and other useful quantities). Functions decorated
  with :py:func:`~.preprocess_plot_grid` automatically gain support for
  :py:class:`~.HierarchicalGridData` and :py:class:`~.UniformGridData`, so they
  only need to worry about plotting NumPy arrays. Functions have to take a
  positional argument ``data`` and a keyword argument ``coordinates``.

Grid data:

- :py:func:`~.plot_color` to plot directly some data with its value.
- :py:func:`~.plot_contourf` to draw a filled contour plot using the data.

Horizons:

- :py:func:`~.plot_horizon` to plot a given shape in 2D.
- :py:func:`~.plot_horizon_on_plane_at_iteration` to plot a given horizon
   at a given iteration in 2D.
- :py:func:`~.plot_horizon_on_plane_at_time` to plot a given horizon
   at a given time in 2D.

Most of the functions here take optional arguments ``figure`` and ``axis``. You
can specify them, or the current ones will be used.

"""

import itertools
import os
import warnings
from typing import Any, Dict, Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.mplot3d import Axes3D

from kuibit import grid_data as gd
from kuibit.cactus_grid_functions import BaseOneGridFunction

# UTILITIES


def setup_matplotlib(
    params: Optional[Dict[str, Any]] = None, rc_par_file: Optional[str] = None
) -> None:
    """Setup matplotlib with some reasonable defaults for better plots.

    If ``params`` is provided, add these parameters to matplotlib's settings
    (``params`` updates ``matplotlib.rcParams``).

    If ``rc_par_file`` is provided, first set the parameters reading the values
    from the ``rc_par_file``. (``params`` has the precedence over the parameters
    read from the file.)

    Matplotlib behaves differently on different machines. With this, we make
    sure that we set all the relevant paramters that we care of to the value we
    prefer. The default values are highly opinionated.

    :param params: Parameters to update matplotlib with.
    :type params: dict

    :param rc_par_file: File where to read parameters. The file has to use
                        matplotlib's configuration language. ``params``
                        overwrites the values set from this file, but this file
                        overrides the default values set in this function.
    :type rc_par_file: str

    """

    matplotlib.rcParams.update(
        {
            "lines.markersize": 4,
            "axes.labelsize": 16,
            "font.weight": "light",
            "font.size": 16,
            "legend.fontsize": 16,
            "xtick.labelsize": 13,
            "ytick.labelsize": 13,
            "axes.formatter.limits": [-3, 3],
            "xtick.minor.visible": True,
            "ytick.minor.visible": True,
            "image.cmap": "inferno",
            "legend.fancybox": False,
            "legend.edgecolor": "inherit",
        }
    )

    if rc_par_file is not None:
        matplotlib.rc_file(rc_par_file)

    if params is not None:
        matplotlib.rcParams.update(params)


def preprocess_plot(func):
    """Decorator to set-up plot functions.

    When we plot anything, there is always some boilerplate that has to be
    executed. For example, we want to provide an axis keyword so that the user
    can specify where to plot, but if the keyword is not provided, we want to
    plot on the current figure.

    Essentially, this decorator sets default values. Why don't we do
    axis=plt.gca() then? The problem is that the default values are set when
    the function is defined, not when it is called. So, this will not work.

    This decorator takes care of everything.

    1. It handles the axis keyword setting it to plt.gca() if it was not
       provided.
    2. It handles the figure keyword setting it to plt.gcf() if it was not
       provided.

    func has to take as keyword arguments:
    1. 'axis=None', where the plot will be plot, or plt.gca() if None
    2. 'figure=None', where the plot will be plot, or plt.gcf() if None

    """

    def inner(*args, **kwargs):
        # Setdetault addes the key if it is not already there
        kwargs.setdefault("axis", plt.gca())
        kwargs.setdefault("figure", plt.gcf())
        return func(*args, **kwargs)

    return inner


def preprocess_plot_grid(func):
    """Decorator to set-up plot functions that plot grid data.

    This decorator extends :py:func:`~.preprocess_plot` for specific functions.

    1. It handles differt types to plot what intuitively one would want to
       plot.
    1a. If the data is a NumPy array with shape 2, just pass the data,
        otherwise raise an error
    1b. If the data is a NumPy array, just pass the data.
    1c. If data is :py:class:`~.UniformGridData`, pass the data and the
        coordinates.
    1d. If data is :py:class:`~.HierarchicalGridData`, read resample it to
        the given grid, then pass do 1c.
    1e. If data is a :py:class:`~.BaseOneGridFunction`, we read the iteration
        and pass to 1d.

    func has to take as keyword arguments (in addition to the ones in
    :py:func`~.preprocess_plot`):
    1. 'data'. data will be passed as a NumPy array, unless it is
               already so.
    2. 'coordinates=None'. coordinates will be passed as a list of NumPy
                           arrays, unless it is not None. Each NumPy
                           array is the coordinates along one axis.

    """

    @preprocess_plot
    def inner(data, *args, **kwargs):
        # The flow is: We check if data is BaseOneGridFunction or derived. If
        # yes, we read the requested iteration. Then, we check if data is
        # HierachicalGridData, if yes, we resample to UniformGridData. Then we
        # work with UniformGridData and handle coordinates, finally we work
        # with NumPy arrays, which is what we pass to the function.

        def attr_not_available(attr):
            """This is a helper function to see if the user passed an attribute
            or if the attribute is None
            """
            return attr not in kwargs or kwargs[attr] is None

        def default_or_kwargs(attr, default):
            """Return default if the attribute is not available in kwargs, otherwise return
            the attribute

            """
            if attr_not_available(attr):
                return default
            return kwargs[attr]

        if isinstance(data, BaseOneGridFunction):
            if attr_not_available("iteration"):
                raise TypeError(
                    "Data has multiple iterations, specify what do you want to plot"
                )

            # Overwrite data with HierarchicalGridData
            data = data[kwargs["iteration"]]

        if isinstance(data, gd.HierarchicalGridData):
            if attr_not_available("shape"):
                raise TypeError(
                    "The data must be resampled but the shape was not provided"
                )

            # If x0 or x1 are None, we use the ones of the grid
            x0 = default_or_kwargs("x0", data.x0)
            x1 = default_or_kwargs("x1", data.x1)
            resample = default_or_kwargs("resample", False)

            # Overwrite data with UniformGridData
            if data.is_masked():
                warnings.warn(
                    "Mask information will be lost with the resampling"
                )

            data = data.to_UniformGridData(
                shape=kwargs["shape"], x0=x0, x1=x1, resample=resample
            )

        if isinstance(data, gd.UniformGridData):
            # We check if the user has passed coordinates too.
            if "coordinates" in kwargs and kwargs["coordinates"] is not None:
                warnings.warn(
                    "Ignoring provided coordinates (data is UniformGridData)."
                    " To specify boundaries, use x0 and x1."
                )

            # If x0 or x1 are None, we use the ones of the grid
            x0 = default_or_kwargs("x0", data.x0)
            x1 = default_or_kwargs("x1", data.x1)
            # If x0 or x1 are provided, then we resample. So, we don't resample
            # only if x0 AND x1 are not provided.
            resampling = not (
                attr_not_available("x0") and attr_not_available("x1")
            )

            if resampling and attr_not_available("shape"):
                raise TypeError(
                    "The data must be resampled but the shape was not provided"
                )

            if resampling:
                resample = default_or_kwargs("resample", False)
                new_grid = gd.UniformGrid(shape=kwargs["shape"], x0=x0, x1=x1)

                if data.is_masked():
                    warnings.warn(
                        "Mask information will be lost with the resampling"
                    )

                data = data.resampled(
                    new_grid, piecewise_constant=(not resample)
                )

            kwargs["coordinates"] = data.coordinates_from_grid()
            # Overwrite data with NumPy array
            data = data.data_xyz

        if isinstance(data, np.ndarray) and data.ndim != 2:
            raise ValueError("Only 2-dimensional data can be plotted")

        # TODO: Check that coordinates are compatible with data

        # We remove what we don't need from kwargs, so that it is not
        # accidentally passed to the function
        def remove_attributes(*attributes):
            for attr in attributes:
                if attr in kwargs:
                    del kwargs[attr]

        remove_attributes("shape", "x0", "x1", "iteration", "resample")

        return func(data, *args, **kwargs)

    return inner


def _process_anchor_info(anchor, offset):
    """Prepare the correct arguments for text_function in
    :py:func:`~.add_text_to_corner`

    :param anchor: Where to place the text? This is defined
                 via cardinal points (e.g., NW for top left).
    :type anchor: str
    :param offset: How far from the edge to put the text? In percentage
                   of the entire figure.
    :type offset: float

    :returns: Horizontal position, vertical position, horizontal alignment
             vertical alignment.
    :rtype: tuple

    """
    # This function has been split off add_text_to_corner only because it is
    # easier to test this way.
    possible_combinations = [
        n1 + n2 for n1, n2 in itertools.product("NSWE", repeat=2) if n1 != n2
    ] + ["N", "S", "W", "E"]

    if anchor not in possible_combinations:
        raise ValueError("Given anchor is invalid. Use cardinal points.")

    # Now we have to parse the cardinal points to find where
    # to put the

    ver_al = None
    ver_pos = 0.5
    if "S" in anchor:
        ver_al = "bottom"
        ver_pos = offset
    elif "N" in anchor:
        ver_al = "top"
        ver_pos = 1 - offset

    hor_al = None
    hor_pos = 0.5
    if "E" in anchor:
        hor_al = "right"
        hor_pos = 1 - offset
    elif "W" in anchor:
        hor_al = "left"
        hor_pos = offset

    return hor_pos, ver_pos, hor_al, ver_al


@preprocess_plot
def add_text_to_corner(text, anchor="SE", figure=None, axis=None, offset=0.01):
    """Add text to a figure.

    Specify the location of the label using cardinal points (NSWE).
    For example, SE is bottom right.

    :param anchor: Where to place the text? This is defined
                 via cardinal points (e.g., NW for top left).
    :type anchor: str
    :param offset: How far from the edge to put the text? In percentage
                   of the entire figure.
    :type offset: float
    """

    # text_function is the correct function to use depending if we are working
    # with 2D or 3D plots.

    text_function = axis.text2D if isinstance(axis, Axes3D) else axis.text

    hor_pos, ver_pos, hor_al, ver_al = _process_anchor_info(anchor, offset)

    kwargs = {}
    if hor_al is not None:
        kwargs.update({"horizontalalignment": hor_al})

    if ver_al is not None:
        kwargs.update({"verticalalignment": ver_al})

    return text_function(
        hor_pos, ver_pos, text, transform=figure.transFigure, **kwargs
    )


@preprocess_plot
def save(
    outputpath,
    figure=None,
    axis=None,
    tikz_clean_figure=False,
    **kwargs,
):
    """Save figure to the given location.

    If the file extension is ``.tikz``, the file will be saved with
    ``tikzplotlib``.

    Unknown arguments are passed to the ``matplotlib.savefig`` or
    ``tikzplotlib.save`` (depending on the extension). In this second case, if
    ``tikz_clean_figure = True``, unknown arguments are first passed to
    ``tikzplotlib.clean_figure``.

    :param outputpath:  Output path with or without extension. If the
                        extension is ``.tikz``, the file is saved with
                        ``tikzplotlib``.
    :type outputpath:  str
    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param tikz_clean_figure: If ``tikzplotlib`` is begin used, reduce the size
                              of the output ``tikz`` file. When this is set to
                              True, unknown arguments are first passed to
                              ``tikzplotlib.clean_figure``, then to
                              ``tikzplotlib.save``. ``tikzplotlib.clean_figure``
                              will change the given figure.
    :type tikz_clean_figure: bool

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    """
    if os.path.splitext(outputpath)[-1] == ".tikz":
        import tikzplotlib

        # If clean_figure is True, we extract from kwargs those argument
        # that tikzplotlib.clean_figure would take. For this, we need to
        # know what argument that function takes.
        #
        # Form https://stackoverflow.com/a/40363565
        if tikz_clean_figure:
            args_names = tikzplotlib.clean_figure.__code__.co_varnames[
                : tikzplotlib.clean_figure.__code__.co_argcount
            ]

            kwargs_clean_figure = {}

            # We split kwargs in those that are supposed to be passed to
            # clean_figure and those that have to be passed to save. For this,
            # we iterate over the arguments taken by clean_figure, if they are
            # in kwargs, then we move them to a new dictionary
            for arg in args_names:
                if arg in kwargs:
                    kwargs_clean_figure.update({arg: kwargs[arg]})
                    del kwargs[arg]

            tikzplotlib.clean_figure(fig=figure, **kwargs_clean_figure)

        tikzplotlib.save(outputpath, **kwargs)
    else:
        figure.savefig(outputpath, **kwargs)


@preprocess_plot
def save_from_dir_filename_ext(
    output_dir,
    file_name,
    file_ext,
    figure=None,
    axis=None,
    tikz_clean_figure=False,
    **kwargs,
):
    """Save figure to a location defined by a folder, a name, and an extension.

    If the ``file_ext`` is ``tikz``, the file will be saved with
    ``tikzplotlib``.

    Unknown arguments are passed to the ``matplotlib.savefig`` or
    ``tikzplotlib.save`` (depending on the extension).

    :param output_dir: Path of a directory where to save the figure
    :type output_dir: str
    :param file_name: Name of the file.
    :type file_name: str
    :param file_extension: Extension of the file, with or without dot.
    :type file_extension: str
    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    """
    ext = file_ext if file_ext[0] == "." else f".{file_ext}"
    outputpath = os.path.join(output_dir, f"{file_name}{ext}")
    return save(
        outputpath,
        figure=figure,
        axis=axis,
        tikz_clean_figure=tikz_clean_figure,
        **kwargs,
    )


@preprocess_plot
def set_axis_limits(
    xmin=None, xmax=None, ymin=None, ymax=None, figure=None, axis=None
):
    """Set limits on the two axes of axis.

    :param xmin: Minimum on the horizontal axis.
    :type xmin: float
    :param xmax: Maximum on the horizontal axis.
    :type xmax: float
    :param ymin: Minimum on the horizontal axis.
    :type ymin: float
    :param ymax: Maximum on the vertical axis.
    :type ymax: float

    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    """
    axis.set(xlim=(xmin, xmax), ylim=(ymin, ymax))


@preprocess_plot
def set_axis_limits_from_args(args, figure=None, axis=None):
    """Set limits on the two axes of axis with data read from ``args``.

    It uses the ``xmin``, ``xmax``, ``ymin``, ``ymax`` attributes.

    :param args: Options provided by the user.
    :type args: `argparse.Namespace`

    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    """
    set_axis_limits(
        xmin=args.xmin,
        xmax=args.xmax,
        ymin=args.ymin,
        ymax=args.ymax,
        figure=figure,
        axis=axis,
    )


def get_figname(args, default):
    """Return the figure name checking if the user has passed one.

    If it is defined, return ``args.figname``, otherwise return default.

    :param args: Options provided by the user.
    :type args: `argparse.Namespace`
    :param default: Default name if ``figname`` is not in ``args``.
    :type default: str

    :returns: Name of the output figure.
    :rtype: str

    """
    if args.figname is None:
        return default
    # figname is not None
    return args.figname


# GRID FUNCTIONS


def _vmin_vmax_extend(data, vmin=None, vmax=None):
    """Helper function to decide what to do with the colorbar (to extend it or not?)."""

    colorbar_extend = "neither"

    if vmin is None:
        vmin = data.min()

    if data.min() < vmin:
        colorbar_extend = "min"

    if vmax is None:
        vmax = data.max()

    if data.max() > vmax:
        if colorbar_extend == "min":
            colorbar_extend = "both"
        else:
            colorbar_extend = "max"

    return vmin, vmax, colorbar_extend


# All the difficult stuff is in preprocess_plot_grid
@preprocess_plot_grid
def _plot_grid(
    data,
    plot_type="color",
    figure=None,
    axis=None,
    coordinates=None,
    xlabel=None,
    ylabel=None,
    colorbar=False,
    label=None,
    logscale=False,
    vmin=None,
    vmax=None,
    aspect_ratio="equal",
    **kwargs,
):
    """Backend of the :py:func:`~.plot_color` and similar functions.

    The type of plot is specified by the variable ``plot_type``.

    Unknown arguments are passed to
    ``imshow`` if plot is color
    ``contourf`` if plot is contourf.
    ``contour`` if plot is contour.

    :param plot_type: Type of plot. It can be: 'color', 'contourf', 'contour'.
    :type plot_type: str
    """

    _known_plot_types = ("color", "contourf", "contour")

    if plot_type not in _known_plot_types:
        raise ValueError(
            f"Unknown plot_type {plot_type} (Options available {_known_plot_types})"
        )

    # Considering all the effort put in preprocess_plot_grid, we we can plot
    # as we were plotting normal NumPy arrays.

    if logscale:
        # We mask the values that are smaller or equal than 0
        data = np.ma.log10(data)

    vmin, vmax, colorbar_extend = _vmin_vmax_extend(data, vmin=vmin, vmax=vmax)

    # To implement vmin and vmax, we clamp the data to vmin and vmax instead of
    # using the options in matplotlib. This greatly simplifies handling things
    # like colormaps.
    data = np.clip(data, vmin, vmax)

    if aspect_ratio is not None:
        axis.set_aspect(aspect_ratio)

    if xlabel is not None:
        axis.set_xlabel(xlabel)
    if ylabel is not None:
        axis.set_ylabel(ylabel)

    if plot_type == "color":
        if coordinates is None:
            grid = None
        else:
            # We assume equally-spaced points.

            # TODO: (Refactoring)
            #
            # This is not a very pythonic way to write this...

            X, Y = coordinates

            dx, dy = X[1] - X[0], Y[1] - Y[0]
            grid = [
                X[0] - 0.5 * dx,
                X[-1] + 0.5 * dx,
                Y[0] - 0.5 * dy,
                Y[-1] + 0.5 * dy,
            ]

        image = axis.imshow(
            data, vmin=vmin, vmax=vmax, origin="lower", extent=grid, **kwargs
        )
    elif plot_type == "contourf":
        if coordinates is None:
            raise ValueError(
                f"You must provide the coordiantes with plot_type = {plot_type}"
            )
        image = axis.contourf(
            *coordinates, data, extend=colorbar_extend, **kwargs
        )
    elif plot_type == "contour":
        if coordinates is None:
            raise ValueError(
                f"You must provide the coordiantes with plot_type = {plot_type}"
            )
        # We need to pass the levels for the contours
        if "levels" not in kwargs:
            raise ValueError(
                f"You must provide the levels with plot_type = {plot_type}"
            )
        image = axis.contour(
            *coordinates, data, extend=colorbar_extend, **kwargs
        )

    if colorbar:
        plot_colorbar(image, figure=figure, axis=axis, label=label)

    return image


def plot_contourf(data, **kwargs):
    """Plot the given data drawing filled contours.

    You can pass (everything is processed by :py:func:`~.preprocess_plot_grid` so
    that at the end we have a 2D NumPy array):
    - A 2D NumPy array,
    - A :py:class:`~.UniformGridData`,
    - A :py:class:`~.HierarchicalGridData`,
    - A :py:class:`~.BaseOneGridFunction`.

    Depending on what you pass, you might need additional arguments.

    If you pass a :py:class:`~.BaseOneGridFunction`, you need also to pass
    ``iteration``, and ``shape``. If you pass
    :py:class:`~.HierarchicalGridData`, you also need to pass ``shape``. In all
    cases you can also pass ``x0`` and ``x1`` to define origin and corner of the
    grid. You can pass the option ``resample=True`` if you want to do bilinear
    resampling at the grid data level, otherwise, nearest neighbor resampling is
    done. When you pass the NumPy array, you also have to pass the
    ``coordinates``.

    All the unknown arguments are passed to ``contourf``.

    .. note

       Read the documentation for a concise table on what arguments are
       supported.

    :param data: Data that has to be plotted. The function expects a 2D NumPy
                 array, but the decorator :py:func:`~.preprocess_plot_grid`
                 allows it to take different kind of data.
    :type data: 2D NumPy array, or object that can be cast to 2D NumPy array.

    :param x0: Lowermost leftmost coordinate to plot. If passed, resampling will
               be performed.
    :type x0: 2D array or list

    :param x1: Uppermost rightmost coordinate to plot. If passed, resampling will
               be performed.
    :type x1: 2D array or list

    :param coordiantes: Coordinates to use for the plot. Used only if data is a
                        NumPy array.
    :type coordinates: 2D array or list

    :param shape: Resolution of the image. This parameter is used if resampling
                  is needed or requested.
    :type shape: tuple or list

    :param iteration: Iteration to plot. Relevant only if data is a
                      :py:class:`~.BaseOneGridData`.
    :type iteration: int

    :param resample: If resampling has to be done, do bilinear resampling at the
                     level of the grid data. If not passed, use nearest neighbors.
    :type resample: bool

    :param logscale: If True, take the log10 of the data before plotting.
    :type logscale: bool

    :param colorbar: If True, add a colorbar.
    :type colorbar: bool

    :param vmin: Remove all the data below this value. If logscale, this has to
                 be the log10.
    :type vmin: float
    :param vmax: Remove all the data above this value. If logscale, this has to
                 be the log10.
    :type vmax: float

    :param xlabel: Label of the x axis. If None (or not passed), no label is
                   placed.
    :type xlabel: str

    :param ylabel: Label of the y axis. If None (or not passed), no label is
                   placed.
    :type ylabel: str

    :param aspect_ratio: Aspect ratio of the plot, as passed to the function
                         ``set_aspect_ratio`` in matplotlib.
    :type aspect_ratio: str

    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    :param kwargs: All the unknown arguments are passed to ``imshow``.
    :type kwargs: dict

    """
    # This function is a convinence function around _plot_grid.
    return _plot_grid(data, plot_type="contourf", **kwargs)


def plot_color(data, **kwargs):
    """Plot the given data.

    You can pass (everything is processed by :py:func:`~.preprocess_plot_grid` so
    that at the end we have a 2D NumPy array):
    - A 2D NumPy array,
    - A :py:class:`~.UniformGridData`,
    - A :py:class:`~.HierarchicalGridData`,
    - A :py:class:`~.BaseOneGridFunction`.

    Depending on what you pass, you might need additional arguments.

    If you pass a :py:class:`~.BaseOneGridFunction`, you need also to pass
    ``iteration``, and ``shape``. If you pass
    :py:class:`~.HierarchicalGridData`, you also need to pass ``shape``. In all
    cases you can also pass ``x0`` and ``x1`` to define origin and corner of the
    grid. You can pass the option ``resample=True`` if you want to do bilinear
    resampling at the grid data level, otherwise, nearest neighbor resampling is
    done. When you pass the NumPy array, passing ``coordinates`` will argument
    will make sure that those coordinates are used.

    All the unknown arguments are passed to ``imshow``.

    .. note

       Read the documentation for a concise table on what arguments are
       supported.

    :param data: Data that has to be plotted. The function expects a 2D NumPy
                 array, but the decorator :py:func:`~.preprocess_plot_grid`
                 allows it to take different kind of data.
    :type data: 2D NumPy array, or object that can be cast to 2D NumPy array.

    :param x0: Lowermost leftmost coordinate to plot. If passed, resampling will
               be performed.
    :type x0: 2D array or list

    :param x1: Uppermost rightmost coordinate to plot. If passed, resampling will
               be performed.
    :type x1: 2D array or list

    :param coordiantes: Coordinates to use for the plot. Used only if data is a
                        NumPy array.
    :type coordinates: 2D array or list

    :param shape: Resolution of the image. This parameter is used if resampling
                  is needed or requested.
    :type shape: tuple or list

    :param iteration: Iteration to plot. Relevant only if data is a
                      :py:class:`~.BaseOneGridData`.
    :type iteration: int

    :param resample: If resampling has to be done, do bilinear resampling at the
                     level of the grid data. If not passed, use nearest neighbors.
    :type resample: bool

    :param logscale: If True, take the log10 of the data before plotting.
    :type logscale: bool

    :param colorbar: If True, add a colorbar.
    :type colorbar: bool

    :param vmin: Remove all the data below this value. If logscale, this has to
                 be the log10.
    :type vmin: float
    :param vmax: Remove all the data above this value. If logscale, this has to
                 be the log10.
    :type vmax: float

    :param xlabel: Label of the x axis. If None (or not passed), no label is
                   placed.
    :type xlabel: str

    :param ylabel: Label of the y axis. If None (or not passed), no label is
                   placed.
    :type ylabel: str

    :param aspect_ratio: Aspect ratio of the plot, as passed to the function
                         ``set_aspect_ratio`` in matplotlib.
    :type aspect_ratio: str

    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    :param kwargs: All the unknown arguments are passed to ``imshow``.
    :type kwargs: dict

    """
    # This function is a convinence function around _plot_grid.
    return _plot_grid(data, plot_type="color", **kwargs)


def plot_contour(data, levels=5, **kwargs):
    """Plot the given data drawing the contours.

    You can pass (everything is processed by :py:func:`~.preprocess_plot_grid` so
    that at the end we have a 2D NumPy array):
    - A 2D NumPy array,
    - A :py:class:`~.UniformGridData`,
    - A :py:class:`~.HierarchicalGridData`,
    - A :py:class:`~.BaseOneGridFunction`.

    Depending on what you pass, you might need additional arguments.

    If you pass a :py:class:`~.BaseOneGridFunction`, you need also to pass
    ``iteration``, and ``shape``. If you pass
    :py:class:`~.HierarchicalGridData`, you also need to pass ``shape``. In all
    cases you can also pass ``x0`` and ``x1`` to define origin and corner of the
    grid. You can pass the option ``resample=True`` if you want to do bilinear
    resampling at the grid data level, otherwise, nearest neighbor resampling is
    done. When you pass the NumPy array, you also have to pass the
    ``coordinates``.

    ``levels`` can be an integer (the number of levels), or an array with the
    specific values where to put the levels.

    All the unknown arguments are passed to ``contour``.

    .. note

       Read the documentation for a concise table on what arguments are
       supported.

    :param data: Data that has to be plotted. The function expects a 2D NumPy
                 array, but the decorator :py:func:`~.preprocess_plot_grid`
                 allows it to take different kind of data.
    :type data: 2D NumPy array, or object that can be cast to 2D NumPy array.

    :param x0: Lowermost leftmost coordinate to plot. If passed, resampling will
               be performed.
    :type x0: 2D array or list

    :param x1: Uppermost rightmost coordinate to plot. If passed, resampling will
               be performed.
    :type x1: 2D array or list

    :param coordiantes: Coordinates to use for the plot. Used only if data is a
                        NumPy array.
    :type coordinates: 2D array or list

    :param shape: Resolution of the image. This parameter is used if resampling
                  is needed or requested.
    :type shape: tuple or list

    :param iteration: Iteration to plot. Relevant only if data is a
                      :py:class:`~.BaseOneGridData`.
    :type iteration: int

    :param resample: If resampling has to be done, do bilinear resampling at the
                     level of the grid data. If not passed, use nearest neighbors.
    :type resample: bool

    :param logscale: If True, take the log10 of the data before plotting.
    :type logscale: bool

    :param colorbar: If True, add a colorbar.
    :type colorbar: bool

    :param vmin: Remove all the data below this value. If logscale, this has to
                 be the log10.
    :type vmin: float
    :param vmax: Remove all the data above this value. If logscale, this has to
                 be the log10.
    :type vmax: float

    :param xlabel: Label of the x axis. If None (or not passed), no label is
                   placed.
    :type xlabel: str

    :param ylabel: Label of the y axis. If None (or not passed), no label is
                   placed.
    :type ylabel: str

    :param levels: If int, the number of levels, if array, the specific levels
                   where to place the contour lines.
    :type levels: int or list

    :param aspect_ratio: Aspect ratio of the plot, as passed to the function
                         ``set_aspect_ratio`` in matplotlib.
    :type aspect_ratio: str

    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    :param kwargs: All the unknown arguments are passed to ``imshow``.
    :type kwargs: dict

    """
    # This function is a convinence function around _plot_grid.
    return _plot_grid(data, plot_type="contour", levels=levels, **kwargs)


@preprocess_plot
def plot_colorbar(
    mpl_artist,
    figure=None,
    axis=None,
    label=None,
    where="right",
    size="5%",
    pad=0.25,
    **kwargs,
):
    """Add a colorbar to an existing image.

    :param mpl_artist: Image from which to generate the colorbar.
    :type mpl_artist: ``matplotlib.cm.ScalarMappable``
    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``
    :param label: Label to place near the colorbar.
    :type label: str
    :param where: Where to place the colorbar (left, right, bottom, top).
    :type where: str
    :param size: Width of the colorbar with respect to ``axis``.
    :type size: float
    :param pad: Pad between the colorbar and ``axis``.
    :type pad: float

    """
    # The next two lines guarantee that the colorbar is the same size as
    # the plot. From https://stackoverflow.com/a/18195921
    divider = make_axes_locatable(axis)
    cax = divider.append_axes(where, size=size, pad=pad)
    cb = plt.colorbar(mpl_artist, cax=cax, **kwargs)
    if label is not None:
        cb.set_label(label)

    # When we draw a colorbar, that changes the selected axis. We do not
    # want that, so we select back the original one.
    plt.sca(axis)

    return cb


@preprocess_plot
def plot_components_boundaries(
    hierarchical_data, figure=None, axis=None, remove_ghosts=True, **kwargs
):
    """Plot the boundaries of all the components available in the data.

    If the components in a given refinement levels can be merged into a single
    one, they be.

    By default, the grids are plotted with black boundaries. This can be
    customized passing the ``edgecolor`` argument.

    :param hierarchical_data: 2D :py:class:`~.HierarchicalGridData` from which
                              to extract the grid structure.
    :type hierarchical_data: :py:class:`~.HierarchicalGridData`,

    :param remove_ghosts: If True, ghosts zones are not included in the plotted
                          grids.
    :type remove_ghosts: bool

    :param kwargs: All the unknown arguments are passed to ``Rectangle``.
    :type kwargs: dict

    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``

    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``

    """
    if not isinstance(hierarchical_data, gd.HierarchicalGridData):
        raise TypeError("The input has to be a HierarchicalGridData")

    if hierarchical_data.num_dimensions != 2:
        raise ValueError("Only 2D HierarchicalGridData can be plotted")

    # Add default color
    if "edgecolor" not in kwargs:
        kwargs["edgecolor"] = "black"

    for _1, _2, comp in hierarchical_data:
        # grid is the UniformGrid of the component under consideration with or
        # without ghost zones depending on the value of remove_ghosts
        grid = comp.grid.ghost_zones_removed() if remove_ghosts else comp.grid

        # comp.highest_vertex and comp.lowest_vertex are 2D NumPy arrays, so
        # comp.highest_vertex - comp.lowest_vertex is the length of the
        # component along the two directions. We unpack it to width and height.
        # We need vertices as opposed to x0 and x1 because we want to take into
        # account the size of the boundary cells
        width, height = grid.highest_vertex - grid.lowest_vertex
        axis.add_patch(
            Rectangle(
                grid.lowest_vertex, width, height, facecolor="none", **kwargs
            )
        )


# HORIZONS


@preprocess_plot
def plot_horizon(
    shape,
    color=None,
    edgecolor=None,
    alpha=None,
    figure=None,
    axis=None,
    **kwargs,
):
    """Plot outline of horizon in 2D.

    Unknown arguments are passed to the ``fill`` function.

    :param shape: Shape of the horizon as returned by
                  `~.shape_outline_at_iteration` or `~.shape_outline_at_time`.
    :type shape: two NumPy arrays
    :param color: Color of the interior of the horizon.
    :type color: color as supported by Matplotlib
    :param edgecolor: Color of the edge of the horizon.
    :type edgecolor: color as supported by Matplotlib
    :param alpha:  Opacity of the horizon.
    :type alpha:  float

    """
    return axis.fill(
        *shape, color=color, edgecolor=edgecolor, alpha=alpha, **kwargs
    )


@preprocess_plot
def _plot_horizon_on_plane(
    horizon,
    plot_type="iteration",
    iteration=None,
    time=None,
    plane="xy",
    time_tolerance=1e-10,
    color=None,
    edgecolor=None,
    alpha=None,
    figure=None,
    axis=None,
    **kwargs,
):
    """Backend for :py:func:`~.plot_horizon_on_plane_at_iteration` and
    :py:func:`~.plot_horizon_on_plane_at_time`.

    Unknown arguments are passed to :py:func:`~.plot_horizon`.

    :param horizon: Horizon to plot.
    :type horizon: :py:class:`~.OneHorizon`
    :param iteration: Iteration to plot.
    :type iteration: int
    :param plane: Plane where to plot (options: `xy`, `xz`, `yz`)
    :type plane: str
    :param time: Time to plot.
    :type time: float
    :param time_tolerance: Tolerance in the determination of the time.
    :type time_tolerance: float
    :param color: Color of the interior of the horizon.
    :type color: color as supported by Matplotlib
    :param edgecolor: Color of the edge of the horizon.
    :type edgecolor: color as supported by Matplotlib
    :param alpha:  Opacity of the horizon.
    :type alpha:  float
    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``
    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``
    """

    cut = {
        "xy": (None, None, 0),
        "xz": (None, 0, None),
        "yz": (0, None, None),
    }

    if plane not in cut:
        raise ValueError(f"Plane has to be one of {list(cut.keys())}")

    if plot_type == "iteration":
        shape = horizon.shape_outline_at_iteration(iteration, cut[plane])
    elif plot_type == "time":
        shape = horizon.shape_outline_at_time(
            time, cut[plane], tolerance=time_tolerance
        )

    if shape is None:
        raise RuntimeError(
            "No outline found on given plane. "
            "This might be due to lack of shape interpolation"
        )

    return plot_horizon(
        shape,
        color=color,
        edgecolor=edgecolor,
        alpha=alpha,
        figure=figure,
        axis=axis,
        **kwargs,
    )


@preprocess_plot
def plot_horizon_on_plane_at_iteration(
    horizon,
    iteration,
    plane,
    color=None,
    edgecolor=None,
    alpha=None,
    figure=None,
    axis=None,
    **kwargs,
):
    """Plot outline of horizon in 2D on a given plane at a given iteration.

    Unknown arguments are passed to :py:func:`~.plot_horizon`.

    .. warning::

       When you take a cross section (an outline) of an horizon, ``kuibit``
       finds points that are within a threshold to the plane that cuts the
       surface. However, the way points are distributed on apparent horizons is
       highly non-uniform. So, if you are cutting the horizon along an axis that
       is not one of the coordinate ones (for the horizon), it is likely that
       too few points will be close enough to the intersecting plane, resulting
       in a malformed or absent outline. In some distant future, ``kuibit`` will
       perform interpolations to solve this problem.

    :param horizon: Horizon to plot.
    :type horizon: :py:class:`~.OneHorizon`
    :param iteration: Iteration to plot.
    :type iteration: int
    :param plane: Plane where to plot (options: `xy`, `xz`, `yz`)
    :type plane: str
    :param color: Color of the interior of the horizon.
    :type color: color as supported by Matplotlib
    :param edgecolor: Color of the edge of the horizon.
    :type edgecolor: color as supported by Matplotlib
    :param alpha:  Opacity of the horizon.
    :type alpha:  float
    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``
    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``
    """
    return _plot_horizon_on_plane(
        horizon,
        plot_type="iteration",
        plane=plane,
        iteration=iteration,
        color=color,
        edgecolor=edgecolor,
        alpha=alpha,
        figure=figure,
        axis=axis,
        **kwargs,
    )


@preprocess_plot
def plot_horizon_on_plane_at_time(
    horizon,
    time,
    plane,
    time_tolerance=1e-10,
    color=None,
    edgecolor=None,
    alpha=None,
    figure=None,
    axis=None,
    **kwargs,
):
    """Plot outline of horizon in 2D on a given plane at a given time.

    Unknown arguments are passed to :py:func:`~.plot_horizon`.

    .. warning::

       When you take a cross section (an outline) of an horizon, ``kuibit``
       finds points that are within a threshold to the plane that cuts the
       surface. However, the way points are distributed on apparent horizons is
       highly non-uniform. So, if you are cutting the horizon along an axis that
       is not one of the coordinate ones (for the horizon), it is likely that
       too few points will be close enough to the intersecting plane, resulting
       in a malformed or absent outline. In some distant future, ``kuibit`` will
       perform interpolations to solve this problem.

    :param horizon: Horizon to plot.
    :type horizon: :py:class:`~.OneHorizon`
    :param time: Time to plot.
    :type time: float
    :param time_tolerance: Tolerance in the determination of the time.
    :type time_tolerance: float
    :param plane: Plane where to plot (options: `xy`, `xz`, `yz`)
    :type plane: str
    :param color: Color of the interior of the horizon.
    :type color: color as supported by Matplotlib
    :param edgecolor: Color of the edge of the horizon.
    :type edgecolor: color as supported by Matplotlib
    :param alpha:  Opacity of the horizon.
    :type alpha:  float
    :param figure: If passed, plot on this figure. If not passed (or if None),
                   use the current figure.
    :type figure: ``matplotlib.pyplot.figure``
    :param axis: If passed, plot on this axis. If not passed (or if None), use
                 the current axis.
    :type axis: ``matplotlib.pyplot.axis``
    """

    return _plot_horizon_on_plane(
        horizon,
        plane=plane,
        plot_type="time",
        time_tolerance=time_tolerance,
        time=time,
        color=color,
        edgecolor=edgecolor,
        alpha=alpha,
        figure=figure,
        axis=axis,
        **kwargs,
    )



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# Based on code originally developed by Wolfgang Kastaun. This file may contain
# algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/cactus_multipoles.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""This module provides access to data saved by the Multipole thorn.

There are multiple classes defined in this module:

- :py:class`~.MultipolesDir` interfaces with :py:class:`~.SimDir` and organizes the
  data according to the variable available. This is a dictionary-like object with keys
  the variable names.
- :py:class`~.MultipoleAllDets` takes all the files that correspond to a given variable
  and organize them according to the extraction radius.
- :py:class`~.MultipoleOneDet` represents one single extraction radius, for a single
  variable. It is a dictionary-like object with keys the multipolar numbers and values
  the multipolar decomposition represented as represented as :py:class:`~.TimeSeries`
  objects.

These are hierarchical classes, one containing the others, so one typically ends
up with a series of brackets or dots to access the actual data. For example, if
``sim`` is a :py:class:`~.SimDir`, ``sim.multipoles['rho_b'][100][2,2]`` is
(2,2) decomposition of ``rho_b`` at radius 100 represented as
:py:class:`~.TimeSeries`.

"""

import os
import re
from typing import Optional

import h5py
import numpy as np

from kuibit import timeseries
from kuibit.attr_dict import pythonize_name_dict


class MultipoleOneDet:
    """This class collects multipole components of a specific variable
    a given spherical surface.

    Multipoles are tightly connected with gravitational waves, so morally a
    sphere where multipoles are computed is a "detector". Hence, the name
    of the class.

    :py:class:`~.MultipoleOneDet` is a dictionary-like object with components as
    the tuples (l,m) and values the corresponding multipolar decomposition as a
    :py:class:`~.TimeSeries` object. Alternatively, this can also be called
    directly ``multipoleonedet(l,m)``. Iteration is supported and yields tuples
    ``(l, m, data)``, which can be used to loop through all the multipoles
    available.

    The reason we allow for ``l_min`` is to remove those files that are not
    necessary when considering gravitational waves (l<2).

    :ivar dist: Radius of the sphere.
    :vartype dist: float
    :ivar radius: Radius of the sphere.
    :type radius: float
    :ivar l_min: l smaller than ``l_min`` are dropped.
    :type l_min: int
    :ivar available_l: Available l values.
    :type available_l: set
    :ivar available_m: Available m values.
    :ivar available_m: set
    :ivar available_lm: Available ``(l, m)`` values.
    :ivar available_lm: set of tuples
    :ivar missing_lm: Missing (l, m) values to have all from ``l_min`` to ``l_max``.
    :ivar missing_lm: set of tuples

    """

    def __init__(self, dist, data, l_min=0):
        """Constructor.

        :param dist: Radius of the spherical surface.
        :type dist: float
        :param data: List of tuples with the two multipolar numbers and
                     the data as :py:class:`~.TimeSeries`.
        :type data: list of tuple ``(l, m, timeseries)``
        :ivar l_min: l smaller than ``l_min`` are dropped.
        :type l_min: int

        """

        self.dist = float(dist)
        self.radius = self.dist  # This is just an alias

        self.l_min = l_min

        # Associate multipoles to list of timeseries
        multipoles_list_ts = {}

        # Now we populate the multipoles_ts_list dictionary. This is a
        # dictionary with keys (l, m) and with values lists of timeseries. If
        # the key is not already present, we create it with value an empty
        # list, then we append the timeseries to this list
        for mult_l, mult_m, ts in data:  # mult = "multipole"
            if mult_l >= l_min:
                lm_list = multipoles_list_ts.setdefault((mult_l, mult_m), [])
                # This means: multipoles[(mult_t, mult_m)] = []
                lm_list.append(ts)
                # At the end we have:
                # multipoles[(mult_t, mult_m)] = [ts1, ts2, ...]

        # Now self._multipoles is a dictionary in which all the timeseries are
        # collapse in a single one. So it is a straightforward map (l, m) -> ts
        self._multipoles = {
            lm: timeseries.combine_ts(ts)
            for lm, ts in multipoles_list_ts.items()
        }
        self.available_l = sorted(
            {mult_l for mult_l, _ in self._multipoles.keys()}
        )
        self.l_max = max(self.available_l)
        self.available_m = sorted(
            {mult_m for _, mult_m in self._multipoles.keys()}
        )
        self.available_lm = set(self._multipoles.keys())

        # Check if all the (l, m) from l_min to l_max are available
        all_lm = set()
        for mult_l in range(self.l_min, max(self.available_l) + 1):
            for mult_m in range(-mult_l, mult_l + 1):
                all_lm.add((mult_l, mult_m))

        # set subtraction
        self.missing_lm = all_lm - self.available_lm

        # Data is in the format expected by __init__
        self.data = [(lm[0], lm[1], ts) for lm, ts in self._multipoles.items()]

    def copy(self):
        """Return a deep copy.

        :returns: Deep copy of ``self``.
        :rtype: :py:class:`~.MultipoleOneDet`
        """
        return type(self)(self.dist, self.data, self.l_min)

    def crop(self, init: Optional[float] = None, end: Optional[float] = None):
        """Remove all the data before ``init`` and after ``end``.

        If ``init`` or ``end`` are not specified, do not crop on that side.
        """
        for _, _, ts in self.data:
            ts.crop(init=init, end=end)

    def cropped(
        self, init: Optional[float] = None, end: Optional[float] = None
    ):
        """Return a copy where the data is cropped in the provided interval.

        If ``init`` or ``end`` are not specified, do not crop on that side.

        :returns: Copy of ``self`` with data cropped.
        :rtype: :py:class:`~.MultipoleOneDet`
        """
        ret = self.copy()
        ret.crop(init=init, end=end)
        return ret

    def __contains__(self, key):
        return key in self._multipoles

    def __getitem__(self, key):
        return self._multipoles[key]

    def __call__(self, mult_l, mult_m):
        return self[(mult_l, mult_m)]

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (
            self.dist == other.dist and self._multipoles == other._multipoles
        )

    # From Python's docs: In order to conform to the object model, classes that
    # define their own equality method should also define their own hash method,
    # or be unhashable.

    # Since we consider series unhashable, this object also has to be unhashable.
    __hash__ = None

    def __iter__(self):
        for (mult_l, mult_m), ts in sorted(self._multipoles.items()):
            yield mult_l, mult_m, ts

    def __len__(self):
        return len(self._multipoles)

    def keys(self):
        """Return available multipolar numbers.

        :returns: Available multipolar numbers.
        :rtype: dict_keys
        """
        return self._multipoles.keys()

    def __str__(self):
        ret = f"(l, m) available: {list(self.keys())}"
        if self.missing_lm:
            ret += f" (missing: {list(self.missing_lm)})"
        return ret

    def total_function_on_available_lm(
        self, function, *args, l_max=None, **kwargs
    ):
        """Evaluate ``function`` on each multipole and accumulate the result.

        ``total_function_on_available_lm`` will call ``function`` with the
        following arguments:

        ``function(timeseries, mult_l, mult_m, dist, *args, **kwargs)``

        If ``function`` does not need some paramters, it should use take
        the ``*args`` argument to ignore the additional paramters that
        are always passed ``(l, m, r)``.

        Values of l larger than ``l_max`` are ignored.

        This method is used to compute quantities like the total power in
        gravitational waves.

        ``function`` can take additional paramters passed directly from
        ``total_function_on_available_lm`` (e.g. ``pcut`` for FFI).

        :params function: Function that has to be applied on each multipole.
        :type function: callable

        :returns: Sum of function applied to each monopole
        :rtype: return type of function

        """
        # This function is used to compute many quantities with waves (e.g.,
        # total strain, total power emitted, ...)
        # It is a little bit ugly, but it works
        if l_max is None:
            l_max = self.l_max

        if l_max > self.l_max:
            raise ValueError("l max larger than l available")

        # The iterator is increasing in (l, m), so we can start from the first
        # element

        iter_self = iter(self)
        try:
            first_l, first_m, first_det = next(iter_self)
        except StopIteration:
            raise RuntimeError("No multipole moments available")
        if first_l > l_max:
            raise ValueError("l max smaller than all l available")

        result = function(
            first_det, first_l, first_m, self.dist, *args, **kwargs
        )

        for mult_l, mult_m, det in iter_self:
            if mult_l <= l_max:
                result += function(
                    det, mult_l, mult_m, self.dist, *args, **kwargs
                )

        return result


class MultipoleAllDets:
    """This class collects available surfaces with multipole data.

    It is a dictionary-like object with keys the spherical surface radius, and
    values :py:class:`MultipoleOneDet` object. Iteration is supported, sorted by
    ascending radius. You can iterate over all the radii and all the available l
    and m with a nested loop.

    :ivar radii:        Available surface radii.
    :type radii: float
    :ivar r_outer:      Radius of the outermost detector.
    :type r_outer: float
    :ivar l_min: l smaller than l_min are dropped.
    :type l_min: int
    :ivar outermost:    Outermost detector.
    :type outermost: :py:class:`~MultipoleOneDet`
    :ivar available_lm: Available components as tuple (l,m).
    :type available_lm: list of tuples
    :ivar available_l:  List of available l.
    :type available_l: list
    :ivar available_m:  List of available m.
    :type available_m: list

    """

    def __init__(self, data, l_min=0):
        """Constructor.

        :param data: List of tuples with ``(multipole_l, multipole_m,
                     extraction_radius, [timeseries])``, where
                     ``[timeseries]`` is a list of the :py:class:`~.TimeSeries`
                     associated.
        :type data: list of tuples

        """
        self.l_min = l_min

        detectors = {}
        self.available_lm = set()

        # Populate detectors
        # detectors is a dictionary with keys the radii and
        # items that look like ([l, m, timeseries, ...]).
        # We accumulate in the list all the ones for the same
        # radius
        for mult_l, mult_m, radius, ts in data:
            if mult_l >= self.l_min:
                # If we don't have the radius yet, let's create an empty list
                d = detectors.setdefault(radius, [])
                # Add the new values
                d.append((mult_l, mult_m, ts))
                # Tally the available l and m
                self.available_lm.add((mult_l, mult_m))

        self._detectors = {
            radius: MultipoleOneDet(radius, multipoles, self.l_min)
            for radius, multipoles in detectors.items()
        }

        # In Python3 .keys() is not a list
        self.radii = sorted(list(self._detectors.keys()))

        if len(self.radii) > 0:
            self.r_outer = self.radii[-1]
            self.outermost = self._detectors[self.r_outer]
        #
        self.available_l = sorted({mult_l for mult_l, _ in self.available_lm})
        if self.available_l:
            self.l_max = max(self.available_l)
        else:
            self.l_max = None
        self.available_m = sorted({mult_m for _, mult_m in self.available_lm})

        # Alias
        self._dets = self._detectors

        # Data is in the format expected by __init__
        # (multipole_l, multipole_m, extraction_radius, [timeseries])
        self.data = []
        for radius, det in self._dets.items():
            for mult_l, mult_m, ts in det:
                self.data.append((mult_l, mult_m, radius, ts))

    def copy(self):
        """Return a deep copy.

        :returns: Deep copy of ``self``.
        :rtype: :py:class:`~.MultipoleAllDets`
        """
        return type(self)(self.data, self.l_min)

    def has_detector(self, mult_l, mult_m, dist):
        """Check if a given multipole component extracted at a given
        distance is available.

        :param mult_l:     Multipole component l.
        :type mult_l:      int
        :param mult_m:     Multipole component m.
        :type mult_m:      int
        :param dist:  Distance of the detector.
        :type dist:   float

        :returns:     If available or not.
        :rtype:       bool
        """
        if dist in self:
            return (mult_l, mult_m) in self[dist]
        return False

    def __contains__(self, key):
        return key in self._dets

    def __getitem__(self, key):
        return self._dets[key]

    def __iter__(self):
        for r in self.radii:
            yield self[r]

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.radii == other.radii and self._dets == other._dets

    # From Python's docs: In order to conform to the object model, classes that
    # define their own equality method should also define their own hash method,
    # or be unhashable.

    # Since we consider series unhashable, this object also has to be unhashable.
    __hash__ = None

    def __len__(self):
        return len(self._dets)

    def keys(self):
        """Return available extraction radii.

        :returns: Available extraction radii.
        :rtype: dict_keys
        """
        return self._dets.keys()

    def __str__(self):
        ret = f"Avilable radii: {list(self.keys())}\n\n"
        for d in sorted(self.keys()):
            ret += f"At radius {d}, {self._dets[d]}\n"
        return ret


class MultipolesDir:
    """This class provides acces to various types of multipole data in a given
    simulation directory.

    This class is like a dictionary, you can access its values using the
    brackets operator, with values that are :py:mod:`~.MultipoleAllDets`. These
    contain the full multipolar description for all the available radii. Files
    are lazily loaded. If both HDF5 and ASCII are present, HDF5 are preferred.
    There's no attempt to combine the two. Alternatively, you can access
    variables with ``get`` or with ``fields.var_name``.

    """

    def __init__(self, sd):
        """Constructor.

        :param sd: Simulation directory.
        :type sd: :py:class:`~.SimDir`
        """
        # self._vars_*_files are dictionary. For _vars_ascii_files, the keys are
        # the variables and the items are sets of tuples of the form
        # (multipole_l, multipole_m, radius, filename) for text files.
        #
        # For example:
        # self._vars_ascii_files =
        # {'psi4': {(2, 2, 110.69, 'output-0000/mp_Psi4_l_m2_r110.69.asc'),
        #           (2, 2, 110.69, 'output-0001/mp_Psi4_l_m2_r110.69.asc')}}

        self._vars_ascii_files = {}

        # For _vars_h5_files, the keys are still the variables, but values are
        # sets with only the files (and not tuples), since we have to read the
        # content to find the l and m
        #
        # For example
        # self._vars_h5_files =
        # {'psi4': {'output-0000/mp_Psi4.h5', 'output-0001/mp_Psi4.h5'}}
        self._vars_h5_files = {}

        # self._vars is the dictionary where we cache the results. The keys are
        # the variables, the values are the corresponding MultipoleAllDets
        # objects. We fill this with __getitem__
        self._vars = {}

        # First, we need to find the multipole files.
        # There are text files and h5 files
        #
        # We use a regular expressions on the name
        # The structure is like mp_Psi4_l_m2_r110.69.asc
        #
        # Let's understand the regexp:
        # 0. ^ and $ means that we match the entire name
        # 1. We match mp_ followed my the variable name, which is
        #    any combination of characters
        # 2. We match _l with a number
        # 3. We match _m with possibly a minus sign and a number
        # 4. We match _r with any combination of numbers with possibly
        #    dots
        # 5. We possibly match a compression
        rx_ascii = re.compile(
            r"""^
        mp_([a-zA-Z0-9\[\]_]+)
        _l(\d+)
        _m([-]?\d+)
        _r([0-9.]+)
        .asc
        (?:.bz2|.gz)?
        $""",
            re.VERBOSE,
        )

        # For h5 files is easy: it is just the var name
        rx_h5 = re.compile(r"^mp_([a-zA-Z0-9\[\]_]+).h5$")

        for f in sd.allfiles:
            filename = os.path.split(f)[1]
            matched_h5 = rx_h5.match(filename)
            matched_ascii = rx_ascii.match(filename)
            if matched_h5 is not None:
                variable_name = matched_h5.group(1).lower()
                var_list = self._vars_h5_files.setdefault(variable_name, set())
                # We are flagging that this h5
                var_list.add(f)
            elif matched_ascii is not None:
                variable_name = matched_ascii.group(1).lower()
                mult_l = int(matched_ascii.group(2))
                mult_m = int(matched_ascii.group(3))
                radius = float(matched_ascii.group(4))
                var_list = self._vars_ascii_files.setdefault(
                    variable_name, set()
                )
                var_list.add((mult_l, mult_m, radius, f))

        # What pythonize_name_dict does is to make the various variables
        # accessible as attributes, e.g. self.fields.rho
        self.fields = pythonize_name_dict(list(self.keys()), self.__getitem__)

    def __contains__(self, key):
        return str(key).lower() in self.keys()

    # The following are staticmethods because they do not depend on the bound
    # object. Using this decorator we save memory because Python will
    # initialize them only once.

    @staticmethod
    def _multipole_from_textfile(path):
        """Read multipole data from a text file.

        :param path: File to read.
        :type path: str

        :returns: Multipole data.
        :rtype: :py:class:`~.TimeSeries`
        """
        a = np.loadtxt(path, unpack=True, ndmin=2)
        if len(a) != 3:
            raise RuntimeError(f"Wrong format in {path}")
        complex_mp = a[1] + 1j * a[2]
        return timeseries.remove_duplicated_iters(a[0], complex_mp)

    @staticmethod
    def _multipoles_from_h5file(path):
        """Read multipole data from a HDF5 file.

        :param path: File to read.
        :type path: str
        :returns: Multipole data.
        :rtype: :py:class:`~.TimeSeries`
        """
        alldets = []
        # This regex matches : l(number)_m(-number)_r(number)
        fieldname_pattern = re.compile(r"l(\d+)_m([-]?\d+)_r([0-9.]+)")

        try:
            with h5py.File(path, "r") as data:
                # Loop over the groups in the hdf5
                for entry in data.keys():
                    matched = fieldname_pattern.match(entry)
                    if matched:
                        mult_l = int(matched.group(1))
                        mult_m = int(matched.group(2))
                        radius = float(matched.group(3))
                        # Read the actual data
                        a = data[entry][()].T
                        complex_mp = a[1] + 1j * a[2]
                        ts = timeseries.remove_duplicated_iters(
                            a[0], complex_mp
                        )
                        alldets.append((mult_l, mult_m, radius, ts))
        except RuntimeError as exce:
            raise RuntimeError(f"File {data} cannot be processed") from exce

        return alldets

    def _multipoles_from_textfiles(self, mpfiles):
        """Read all the multipole data in several text files.

        :param mpfiles: Files to read.
        :type mpfiles: list of str

        :returns: :py:class:`~.MultipoleAllDets` with all the data read.
        :rtype: :py:class:`~.MultipoleAllDets`
        """
        # We prepare the data for MultipoleAllDets checking for errors
        alldets = [
            (
                mult_l,
                mult_m,
                radius,
                self._multipole_from_textfile(filename),
            )
            for mult_l, mult_m, radius, filename in mpfiles
        ]
        return MultipoleAllDets(alldets)

    def _multipoles_from_h5files(self, mpfiles):
        """Read all the multipole data in several HDF5 files.

        :param mpfiles: Files to read.
        :type mpfiles: list of str

        :returns: :py:class:`~.MultipoleAllDets` with all the data read.
        :rtype: :py:class:`~.MultipoleAllDets`
        """
        mult_l = []

        for filename in mpfiles:
            mult_l.extend(self._multipoles_from_h5file(filename))

        return MultipoleAllDets(mult_l)

    def __getitem__(self, key):
        """Read data associated to variable ``key``.

        HDF5 files are preferred over ASCII ones.

        :returns: Multipolar data.
        :rtype: :py:class:`~.MultipoleAllDets`

        """
        k = str(key).lower()

        # If we don't have already cached the result, we first read it.

        if k not in self._vars:
            # We prefer h5
            if k in self._vars_h5_files:
                self._vars[k] = self._multipoles_from_h5files(
                    self._vars_h5_files[k]
                )
            elif k in self._vars_ascii_files:
                self._vars[k] = self._multipoles_from_textfiles(
                    self._vars_ascii_files[k]
                )
            else:
                raise KeyError

        return self._vars[k]

    def get(self, key, default=None):
        """Return a the multipolar data for the given variable if available, else return
        the default value.

        :param key: Requested variable.
        :type key: str
        :param default: Returned value if ``key`` is not available.
        :type default: any

        :returns: Collection of all the multipolar data for the given variable.
        :rtype: :py:class:`~.MultipoleAllDets`

        """

        if key not in self:
            return default
        return self[key]

    def keys(self):
        """Return available variables with multipolar data.

        :returns: Available variables that have multipolar data.
        :rtype: dict_keys
        """
        # We merge the dictionaries and return the keys.
        # This automatically takes care of making sure that they keys are unique.
        return {**self._vars_h5_files, **self._vars_ascii_files}.keys()

    def __str__(self):
        """NOTE: __str__ requires opening all the h5 files! This can be slow!"""
        ret = f"Variables available: {self.keys()}\n"
        for variable in self.keys():
            ret += f"For variable {variable}:\n\n"
            ret += f"{self[variable]}\n"
        return ret
